"""Implementation of an AbstractFourierOperator. Put simply, these are
functions commonly applied to images in fourier space.

Opposed to a AbstractFilter, a AbstractFourierOperator is computed at
runtime---not upon initialization. AbstractFourierOperators also do not
have a rule for how they should be applied to images.

These classes are modified from the library ``tinygp``.
"""

from abc import abstractmethod
from typing import overload
from typing_extensions import override

import jax.numpy as jnp
from equinox import field
from jaxtyping import Array, Float, Inexact

from ...core import error_if_not_positive
from ._operator import AbstractImageOperator


class AbstractFourierOperator(AbstractImageOperator, strict=True):
    """
    The base class for all fourier-based operators.

    By convention, operators should be defined to
    be dimensionless (up to a scale factor).

    To create a subclass,

        1) Include the necessary parameters in
           the class definition.
        2) Overrwrite the ``__call__`` method.
    """

    @overload
    @abstractmethod
    def __call__(
        self, freqs: Float[Array, "y_dim x_dim 2"]
    ) -> Inexact[Array, "y_dim x_dim"]: ...

    @overload
    @abstractmethod
    def __call__(
        self, freqs: Float[Array, "z_dim y_dim x_dim 3"]
    ) -> Inexact[Array, "z_dim y_dim x_dim"]: ...

    @abstractmethod
    def __call__(  # pyright: ignore
        self, freqs: Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
    ) -> Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"]:
        raise NotImplementedError


FourierOperatorLike = AbstractFourierOperator | AbstractImageOperator


class ZeroMode(AbstractFourierOperator, strict=True):
    """This operator returns a constant in the zero mode."""

    value: Float[Array, ""] = field(default=0.0, converter=jnp.asarray)

    @override
    def __call__(
        self, freqs: Float[Array, "y_dim x_dim 2"]
    ) -> Float[Array, "y_dim x_dim"]:
        N1, N2 = freqs.shape[0:-1]
        return jnp.zeros((N1, N2)).at[0, 0].set(self.value)


ZeroMode.__init__.__doc__ = """**Arguments:**

- `value`: The value of the zero mode.
"""


class FourierExp2D(AbstractFourierOperator, strict=True):
    r"""This operator, in real space, represents a
    function equal to an exponential decay, given by

    .. math::
        g(|r|) = \frac{\kappa}{2 \pi \xi^2} \exp(- |r| / \xi),

    where :math:`|r| = \sqrt{x^2 + y^2}` is a radial coordinate.
    Here, :math:`\xi` has dimensions of length and :math:`g(r)`
    has dimensions of inverse area. The power spectrum from such
    a correlation function (in two-dimensions) is given by its
    Hankel transform pair

    .. math::
        P(|k|) = \frac{\kappa}{2 \pi \xi^3} \frac{1}{(\xi^{-2} + |k|^2)^{3/2}}.

    Here :math:`\kappa` is a scale factor and :math:`\xi` is a length
    scale.
    """

    amplitude: Float[Array, ""] = field(default=1.0, converter=jnp.asarray)
    length_scale: Float[Array, ""] = field(default=1.0, converter=error_if_not_positive)

    @override
    def __call__(
        self, freqs: Float[Array, "y_dim x_dim 2"]
    ) -> Float[Array, "y_dim x_dim"]:
        k_sqr = jnp.sum(freqs**2, axis=-1)
        scaling = 1.0 / (k_sqr + jnp.divide(1, (self.length_scale) ** 2)) ** 1.5
        scaling *= jnp.divide(self.amplitude, 2 * jnp.pi * self.length_scale**3)
        return scaling


FourierExp2D.__init__.__doc__ = """**Arguments:**

- `amplitude`: The amplitude of the operator, equal to $\\kappa$
           in the above equation.
- `length_scale`: The length scale of the operator, equal to $\\xi$
              in the above equation.
"""


class Lorenzian(AbstractFourierOperator, strict=True):
    r"""This operator is the Lorenzian, given

    .. math::
        P(|k|) = \frac{\kappa}{\xi^2} \frac{1}{(\xi^{-2} + |k|^2)}.

    Here :math:`\kappa` is a scale factor and :math:`\xi` is a length
    scale.
    """

    amplitude: Float[Array, ""] = field(default=1.0, converter=jnp.asarray)
    length_scale: Float[Array, ""] = field(default=1.0, converter=error_if_not_positive)

    @overload
    def __call__(
        self, freqs: Float[Array, "y_dim x_dim 2"]
    ) -> Float[Array, "y_dim x_dim"]: ...

    @overload
    def __call__(
        self, freqs: Float[Array, "z_dim y_dim x_dim 3"]
    ) -> Float[Array, "z_dim y_dim x_dim"]: ...

    @override
    def __call__(
        self, freqs: Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
    ) -> Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]:
        k_sqr = jnp.sum(freqs**2, axis=-1)
        scaling = 1.0 / (k_sqr + jnp.divide(1, self.length_scale**2))
        scaling *= jnp.divide(self.amplitude, self.length_scale**2)
        return scaling


Lorenzian.__init__.__doc__ = """**Arguments:**

- `amplitude`: The amplitude of the operator, equal to $\\kappa$
           in the above equation.
- `length_scale`: The length scale of the operator, equal to $\\xi$
              in the above equation.
"""


class FourierGaussian(AbstractFourierOperator, strict=True):
    r"""This operator represents a simple gaussian.
    Specifically, this is

    .. math::
        P(k) = \kappa \exp(- \beta k^2 / 2),

    where :math:`k^2 = k_x^2 + k_y^2` is the length of the
    wave vector. Here, :math:`\beta` has dimensions of length
    squared. In 2D, the real-space version of this function is given
    by

    .. math::
        g(r) = \frac{\kappa}{2\pi \beta} \exp(- r^2 / (2 \beta)),

    where :math:`r^2 = x^2 + y^2`.
    """

    amplitude: Float[Array, ""] = field(default=1.0, converter=jnp.asarray)
    b_factor: Float[Array, ""] = field(default=1.0, converter=error_if_not_positive)

    @overload
    def __call__(
        self, freqs: Float[Array, "y_dim x_dim 2"]
    ) -> Float[Array, "y_dim x_dim"]: ...

    @overload
    def __call__(
        self, freqs: Float[Array, "z_dim y_dim x_dim 3"]
    ) -> Float[Array, "z_dim y_dim x_dim"]: ...

    @override
    def __call__(
        self, freqs: Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]
    ) -> Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]:
        k_sqr = jnp.sum(freqs**2, axis=-1)
        scaling = self.amplitude * jnp.exp(-0.5 * self.b_factor * k_sqr)
        return scaling


FourierGaussian.__init__.__doc__ = """**Arguments:**

- `amplitude`: The amplitude of the operator, equal to $\\kappa$
           in the above equation.
- `b_factor`: The variance of the real-space gaussian, equal to $\\beta$
              in the above equation.
"""



"""
Image formation models simulated from gaussian noise distributions.
"""

from typing import Optional
from typing_extensions import override

import jax.numpy as jnp
import jax.random as jr
import numpy as np
from equinox import field
from jaxtyping import Array, Complex, Float, PRNGKeyArray

from ...core import error_if_not_positive
from ...image.operators import Constant, FourierOperatorLike
from ...simulator import AbstractPipeline
from ._distribution import AbstractDistribution


class IndependentFourierGaussian(AbstractDistribution, strict=True):
    r"""A gaussian noise model, where each fourier mode is independent.

    This computes the likelihood in Fourier space,
    so that the variance to be an arbitrary noise power spectrum.
    """

    pipeline: AbstractPipeline
    variance: FourierOperatorLike
    contrast_scale: Float[Array, ""] = field(converter=error_if_not_positive)

    def __init__(
        self,
        pipeline: AbstractPipeline,
        variance: Optional[FourierOperatorLike] = None,
        contrast_scale: float | Float[Array, ""] = 1.0,
    ):
        """**Arguments:**

        - `pipeline`: The image formation model.
        - `variance`: The variance of each fourier mode. By default,
                      `cryojax.image.operators.Constant(1.0)`.
        - `contrast_scale`: The standard deviation of an image simulated
                            from `pipeline`, excluding the noise. By default,
                            `1.0`.
        """
        self.pipeline = pipeline
        self.variance = variance or Constant(1.0)
        self.contrast_scale = jnp.asarray(contrast_scale)

    @override
    def render(
        self, *, get_real: bool = True
    ) -> (
        Float[Array, "{self.pipeline.config.y_dim} {self.pipeline.config.x_dim}"]
        | Complex[Array, "{self.pipeline.config.y_dim} {self.config.x_dim//2+1}"]
    ):
        """Render the image formation model."""
        return self.contrast_scale * self.pipeline.render(
            normalize=True, get_real=get_real
        )

    @override
    def sample(
        self, key: PRNGKeyArray, *, get_real: bool = True
    ) -> (
        Float[Array, "{self.pipeline.config.y_dim} {self.pipeline.config.x_dim}"]
        | Complex[Array, "{self.pipeline.config.y_dim} {self.config.x_dim//2+1}"]
    ):
        """Sample from the gaussian noise model."""
        N_pix = np.prod(self.pipeline.config.padded_shape)
        freqs = self.pipeline.config.wrapped_padded_frequency_grid_in_angstroms.get()
        # Compute the zero mean variance and scale up to be independent of the number of
        # pixels
        std = jnp.sqrt(N_pix * self.variance(freqs))
        noise = self.pipeline.crop_and_apply_operators(
            std * jr.normal(key, shape=freqs.shape[0:-1]).at[0, 0].set(0.0),
            get_real=get_real,
        )
        image = self.render(get_real=get_real)
        return image + noise

    @override
    def log_likelihood(
        self,
        observed: Complex[
            Array,
            "{self.pipeline.config.y_dim} {self.pipeline.config.x_dim//2+1}",
        ],
    ) -> Float[Array, ""]:
        """Evaluate the log-likelihood of the gaussian noise model.

        **Arguments:**

        - `observed` : The observed data in fourier space.
        """
        N_pix = np.prod(self.pipeline.config.shape)
        freqs = self.pipeline.config.wrapped_frequency_grid_in_angstroms.get()
        # Compute the variance and scale up to be independent of the number of pixels
        variance = N_pix * self.variance(freqs)
        # Create simulated data
        simulated = self.render(get_real=False)
        # Compute residuals
        residuals = simulated - observed
        # Compute standard normal random variables
        squared_standard_normal_per_mode = jnp.abs(residuals) ** 2 / (2 * variance)
        # Compute the log-likelihood for each fourier mode. Divide by the
        # number of pixels so that the likelihood is a sum over pixels in
        # real space (parseval's theorem)
        log_likelihood_per_mode = (
            squared_standard_normal_per_mode - jnp.log(2 * jnp.pi * variance) / 2
        ) / N_pix
        # Compute log-likelihood, throwing away the zero mode. Need to take care
        # to compute the loss function in fourier space for a real-valued function.
        log_likelihood = -1.0 * (
            jnp.sum(log_likelihood_per_mode[1:, 0])
            + 2 * jnp.sum(log_likelihood_per_mode[:, 1:])
        )

        return log_likelihood



"""
Using the fourier slice theorem for computing volume projections.
"""

from typing import Any

import jax.numpy as jnp
from equinox import field
from jaxtyping import Array, Complex, Float

from ...image import (
    irfftn,
    map_coordinates,
    map_coordinates_with_cubic_spline,
    rfftn,
)
from .._config import ImageConfig
from .._potential import (
    FourierVoxelGridPotential,
    FourierVoxelGridPotentialInterpolator,
)
from ._potential_integrator import AbstractPotentialIntegrator


class FourierSliceExtract(AbstractPotentialIntegrator, strict=True):
    """Integrate points to the exit plane using the
    Fourier-projection slice theorem.

    This extracts slices using resampling techniques housed in
    ``cryojax.image._map_coordinates``. See here for more documentation.

    Attributes
    ----------
    interpolation_order :
        The interpolation order. This can be ``0`` (nearest-neighbor), ``1``
        (linear), or ``3`` (cubic).
        Note that this argument is ignored if a ``FourierVoxelGridInterpolator``
        is passed.
    interpolation_mode :
        Specify how to handle out of bounds indexing.
    interpolation_cval :
        Value for filling out-of-bounds indices. Used only when
        ``interpolation_mode = "fill"``.
    """

    interpolation_order: int = field(static=True, default=1)
    interpolation_mode: str = field(static=True, default="fill")
    interpolation_cval: complex = field(static=True, default=0.0 + 0.0j)

    def __call__(
        self,
        potential: FourierVoxelGridPotential | FourierVoxelGridPotentialInterpolator,
        wavelength_in_angstroms: Float[Array, ""],
        config: ImageConfig,
    ) -> Complex[Array, "{config.padded_y_dim} {config.padded_x_dim//2+1}"]:
        """Compute a projection of the real-space potential by extracting
        a central slice in fourier space.
        """
        frequency_slice = potential.wrapped_frequency_slice_in_pixels.get()
        N = frequency_slice.shape[1]
        if potential.shape != (N, N, N):
            raise AttributeError(
                "Only cubic boxes are supported for fourier slice extraction."
            )
        # Compute the fourier projection
        if isinstance(potential, FourierVoxelGridPotentialInterpolator):
            fourier_projection = extract_slice_with_cubic_spline(
                potential.coefficients,
                frequency_slice,
                mode=self.interpolation_mode,
                cval=self.interpolation_cval,
            )
        elif isinstance(potential, FourierVoxelGridPotential):
            fourier_projection = extract_slice(
                potential.fourier_voxel_grid,
                frequency_slice,
                interpolation_order=self.interpolation_order,
                mode=self.interpolation_mode,
                cval=self.interpolation_cval,
            )
        else:
            raise ValueError(
                "Supported density representations are FourierVoxelGrid and "
                "FourierVoxelGridInterpolator."
            )

        # Resize the image to match the ImageConfig.padded_shape
        if config.padded_shape != (N, N):
            fourier_projection = rfftn(
                config.crop_or_pad_to_padded_shape(irfftn(fourier_projection, s=(N, N)))
            )
        # Rescale the voxel size to the ImageConfig.pixel_size
        return config.rescale_to_pixel_size(
            fourier_projection, potential.voxel_size, is_real=False
        )


def extract_slice(
    fourier_voxel_grid: Complex[Array, "dim dim dim"],
    frequency_slice: Float[Array, "1 dim dim 3"],
    interpolation_order: int = 1,
    **kwargs: Any,
) -> Complex[Array, "dim dim//2+1"]:
    """
    Project and interpolate 3D volume point cloud
    onto imaging plane using the fourier slice theorem.

    Arguments
    ---------
    fourier_voxel_grid : shape `(N, N, N)`
        Density grid in fourier space. The zero frequency component
        should be in the center.
    frequency_slice : shape `(1, N, N, 3)`
        Frequency central slice coordinate system, with the zero
        frequency component in the corner.
    interpolation_order : int
        Order of interpolation, either 0, 1, or 3.
    kwargs
        Keyword arguments passed to ``cryojax.image.map_coordinates``
        or ``cryojax.image.map_coordinates_with_cubic_spline``.

    Returns
    -------
    projection : shape `(N, N//2+1)`
        The output image in fourier space.
    """
    # Convert to logical coordinates
    N = frequency_slice.shape[1]
    logical_frequency_slice = (frequency_slice * N) + N // 2
    # Convert arguments to map_coordinates convention and compute
    k_z, k_y, k_x = jnp.transpose(logical_frequency_slice, axes=[3, 0, 1, 2])
    projection = map_coordinates(
        fourier_voxel_grid, (k_x, k_y, k_z), interpolation_order, **kwargs
    )[0, :, :]
    # Shift zero frequency component to corner and take upper half plane
    projection = jnp.fft.ifftshift(projection)[:, : N // 2 + 1]
    # Set last line of frequencies to zero if image dimension is even
    if N % 2 == 0:
        projection = projection.at[:, -1].set(0.0 + 0.0j).at[N // 2, :].set(0.0 + 0.0j)
    return projection


def extract_slice_with_cubic_spline(
    spline_coefficients: Complex[Array, "dim+2 dim+2 dim+2"],
    frequency_slice: Float[Array, "1 dim dim 3"],
    **kwargs: Any,
) -> Complex[Array, "dim dim//2+1"]:
    """
    Project and interpolate 3D volume point cloud
    onto imaging plane using the fourier slice theorem, using cubic
    spline coefficients as input.

    Arguments
    ---------
    spline_coefficients : shape `(N+2, N+2, N+2)`
        Coefficients for cubic spline.
    frequency_slice : shape `(1, N, N, 3)`
        Frequency central slice coordinate system, with the zero
        frequency component in the corner.
    kwargs
        Keyword arguments passed to ``cryojax.image.map_coordinates_with_cubic_spline``.

    Returns
    -------
    projection : shape `(N, N//2+1)`
        The output image in fourier space.
    """
    # Convert to logical coordinates
    N = frequency_slice.shape[1]
    logical_frequency_slice = (frequency_slice * N) + N // 2
    # Convert arguments to map_coordinates convention and compute
    k_z, k_y, k_x = jnp.transpose(logical_frequency_slice, axes=[3, 0, 1, 2])
    projection = map_coordinates_with_cubic_spline(
        spline_coefficients, (k_x, k_y, k_z), **kwargs
    )[0, :, :]
    # Shift zero frequency component to corner and take upper half plane
    projection = jnp.fft.ifftshift(projection)[:, : N // 2 + 1]
    # Set last line of frequencies to zero if image dimension is even
    return projection if N % 2 == 1 else projection.at[:, -1].set(0.0 + 0.0j)



"""
Utilities for runtime errors, wrapping `equinox.error_if`. 
"""

import equinox as eqx
import jax.numpy as jnp
from jaxtyping import Array, ArrayLike


def error_if_negative(x: ArrayLike) -> Array:
    x = jnp.asarray(x)
    return eqx.error_if(x, x < 0, "A non-negative quantity was found to be negative!")


def error_if_not_positive(x: ArrayLike) -> Array:
    x = jnp.asarray(x)
    return eqx.error_if(
        x, x <= 0, "A positive quantity was found to be negative or zero!"
    )


def error_if_zero(x: ArrayLike) -> Array:
    x = jnp.asarray(x)
    return eqx.error_if(
        x, jnp.isclose(x, 0.0), "A non-zero quantity was found to be zero!"
    )


def error_if_not_fractional(x: ArrayLike) -> Array:
    x = jnp.asarray(x)
    return eqx.error_if(
        x,
        ~jnp.logical_and(x >= 0.0, x <= 1.0),
        "A fractional quantity was found to not be between 0 and 1!",
    )



import pytest


@pytest.mark.parametrize("model", ["noisy_model", "filtered_and_masked_model"])
def test_real_shape(model, request):
    """Make sure shapes are as expected in real space."""
    model = request.getfixturevalue(model)
    image = model.render()
    padded_image = model.render(view_cropped=False)
    assert image.shape == model.config.shape
    assert padded_image.shape == model.config.padded_shape


@pytest.mark.parametrize("model", ["noisy_model", "filtered_and_masked_model"])
def test_fourier_shape(model, request):
    """Make sure shapes are as expected in fourier space."""
    model = request.getfixturevalue(model)
    image = model.render(get_real=False)
    padded_image = model.render(view_cropped=False, get_real=False)
    assert image.shape == model.config.wrapped_frequency_grid_in_pixels.get().shape[0:2]
    assert (
        padded_image.shape
        == model.config.wrapped_padded_frequency_grid_in_pixels.get().shape[0:2]
    )



"""
Abstraction of a helical polymer.
"""

from functools import cached_property
from typing import Optional

import jax
import jax.numpy as jnp
from equinox import field
from jaxtyping import Array, Float

from ...rotations import SO3
from .._conformation import AbstractConformation
from .._pose import AbstractPose, EulerAnglePose
from .._specimen import AbstractSpecimen
from ._assembly import AbstractAssembly


class Helix(AbstractAssembly, strict=True):
    """
    Abstraction of a helical polymer.

    This class assembles a helix from a subunit.
    See the ``AbstractAssembly`` base class for more information.

    The screw axis is taken to be in the center of the
    image, pointing out-of-plane (i.e. along the z direction).
    """

    subunit: AbstractSpecimen
    rise: Float[Array, ""]
    twist: Float[Array, ""]

    pose: AbstractPose
    conformation: Optional[AbstractConformation]

    n_subunits: int = field(static=True)
    n_start: int = field(static=True)

    def __init__(
        self,
        subunit: AbstractSpecimen,
        rise: Float[Array, ""] | float,
        twist: Float[Array, ""] | float,
        pose: Optional[AbstractPose] = None,
        conformation: Optional[AbstractConformation] = None,
        n_start: int = 1,
        n_subunits: int = 1,
    ):
        """**Arguments:**
        - `subunit`:
            The helical subunit. It is important to set the the initial pose
            of the initial subunit here. This is in the center of mass frame
            of the helix with the screw axis pointing in the z-direction.
        - `rise`:
            The helical rise. This has dimensions of length.
        - `twist`:
            The helical twist.
        - `pose`:
            The center of mass pose of the helix.
        - `conformation`:
            The conformation of `subunit` at each lattice site.
            This can either be a fixed set of conformations or a function
            that computes conformations based on the lattice positions.
            In either case, the `Array` should be shape
            `(n_start*n_subunits_per_start,)`.
        - `n_start`:
            The start number of the helix. By default, ``1``.
        - `n_subunits_per_start`:
            The number of subunits each helical strand.
            By default, ``1``.
        """
        self.subunit = subunit
        self.pose = pose or EulerAnglePose()
        self.rise = jnp.asarray(rise)
        self.twist = jnp.asarray(twist)
        self.conformation = conformation
        self.n_start = n_start
        self.n_subunits = n_subunits

    def __check_init__(self):
        if self.n_subunits % self.n_start != 0:
            raise AttributeError(
                "The number of subunits must be a multiple of the helical start number."
            )

    @cached_property
    def offsets_in_angstroms(self) -> Float[Array, "{self.n_subunits} 3"]:
        """Get the helical lattice positions in the center of mass frame."""
        return compute_helical_lattice_positions(
            self.rise,
            self.twist,
            n_subunits_per_start=self.n_subunits // self.n_start,
            initial_displacement=self.subunit.pose.offset_in_angstroms,
            n_start=self.n_start,
        )

    @cached_property
    def rotations(self) -> SO3:
        """Get the helical lattice rotations in the center of mass frame.

        These are rotations of the initial subunit.
        """
        transformed_rotation_matrices = compute_helical_lattice_rotations(
            self.twist,
            n_subunits_per_start=self.n_subunits // self.n_start,
            initial_rotation=self.subunit.pose.rotation.as_matrix(),
            n_start=self.n_start,
        )
        # Function to construct SO3 objects vmapped over leading dimension
        transformed_rotations = jax.vmap(lambda mat: SO3.from_matrix(mat))(
            transformed_rotation_matrices
        )
        return transformed_rotations


def compute_helical_lattice_positions(
    rise: Float[Array, ""],
    twist: Float[Array, ""],
    n_subunits_per_start: int,
    initial_displacement: Float[Array, "3"],
    n_start: int = 1,
) -> Float[Array, "{n_start*n_subunits_per_start} 3"]:
    """
    Compute the lattice points of a helix for a given
    rise, twist, radius, and start number.

    Arguments
    ---------
    rise : `Real_` or `RealVector`, shape `(n_subunits,)`
        The helical rise.
    twist : `Real_` or `RealVector`, shape `(n_subunits,)`
        The helical twist.
    n_subunits_per_start :
        The number of subunits in the assembly for
        a single sub-helix.
    initial_displacement : `Array`, shape `(3,)`
        The initial position vector of the first subunit, in
        the center of mass frame of the helix.
        The xy values are an in-plane displacement from
        the screw axis, and the z value is an offset from the
        first subunit's position.
    n_start :
        The start number of the helix.

    Returns
    -------
    subunit_positions : shape `(n_start*n_subunits_per_start, 3)`
        The helical lattice positions.
    """
    # Convert to radians
    twist = jnp.deg2rad(twist)

    # Coordinate transformation to get subunit positions in a single sub-helix
    def compute_ith_subunit_position_in_subhelix(i, theta, dz, r_0, N):
        # ... define rotation about the screw axis
        c, s = jnp.cos(i * theta), jnp.sin(i * theta)
        R_t = jnp.array(((c, -s, 0), (s, c, 0), (0, 0, 1)), dtype=float)
        # ... transform by i rises and i twists
        r_i = R_t @ r_0 + i * jnp.asarray((0, 0, dz), dtype=float)
        # ... center positions of subunits in z
        return r_i - jnp.asarray(
            (0.0, 0.0, dz * jnp.asarray(N - 1, dtype=float) / 2), dtype=float
        )

    # ... function to get positions of all subunits
    compute_subunit_positions_in_subhelix = jax.vmap(
        compute_ith_subunit_position_in_subhelix,
        in_axes=[0, None, None, None, None],
    )
    # ... get indices of subunits along sub-helix
    subunit_indices = jnp.arange(n_subunits_per_start, dtype=float)
    # ... compute positions of subunits in sub-helix
    subunit_positions_in_subhelix = compute_subunit_positions_in_subhelix(
        subunit_indices,
        twist,
        rise,
        initial_displacement,
        n_subunits_per_start,
    )

    # Now, transform this single sub-helix into all sub-helices, related by C_n
    # symmetry
    def compute_helix_subunit_positions_per_start(symmetry_angle, r):
        # ... rotate the sub-helix around the screw axis to a different sub-helix
        c_n, s_n = jnp.cos(symmetry_angle), jnp.sin(symmetry_angle)
        R_n = jnp.array(((c_n, -s_n, 0), (s_n, c_n, 0), (0, 0, 1)), dtype=float)
        return (R_n @ r.T).T

    # ... function to rotate entire sub-helix around the screw axis
    compute_helix_subunit_positions = jax.vmap(
        compute_helix_subunit_positions_per_start, in_axes=[0, None]
    )
    # ... compute symmetry angles relating first sub-helix to all other sub-helices
    symmetry_angles = jnp.array([2 * jnp.pi * n / n_start for n in range(n_start)])
    # ... finally, get all subunit positions!
    subunit_positions = compute_helix_subunit_positions(
        symmetry_angles, subunit_positions_in_subhelix
    ).reshape((n_start * n_subunits_per_start, 3))

    return subunit_positions


def compute_helical_lattice_rotations(
    twist: Float[Array, ""],
    n_subunits_per_start: int,
    initial_rotation: Float[Array, "3 3"] = jnp.eye(3),
    n_start: int = 1,
) -> Float[Array, "{n_start*n_subunits_per_start} 3 3"]:
    """
    Compute the relative rotations of subunits on a
    helical lattice, parameterized by the
    rise, twist, start number, and an initial rotation.

    Arguments
    ---------
    twist : `Real_`
        The helical twist.
    n_subunits_per_start :
        The number of subunits in the assembly for
        a single sub-helix.
    initial_rotation : `Array`, shape `(3, 3)`
        The initial rotation of the first subunit. By default,
        the identity matrix.
    n_start :
        The start number of the helix.

    Returns
    -------
    subunit_rotations : shape `(n_start*n_subunits_per_start, 3, 3)`
        The relative rotations between subunits
        on the helical lattice.
    """
    # Convert to radians
    twist = jnp.deg2rad(twist)

    # Coordinate transformation to get subunit rotations in a single sub-helix
    def compute_ith_subunit_rotation_in_subhelix(i, theta, R_0):
        # ... define rotation about the screw axis
        c, s = jnp.cos(i * theta), jnp.sin(i * theta)
        R_t = jnp.array(((c, -s, 0), (s, c, 0), (0, 0, 1)), dtype=float)
        # ... transform to the ith position
        return R_t @ R_0

    # ... function to get rotations of all subunits
    compute_subunit_rotations_in_subhelix = jax.vmap(
        compute_ith_subunit_rotation_in_subhelix, in_axes=[0, None, None]
    )
    # ... get indices of subunits along sub-helix
    subunit_indices = jnp.arange(n_subunits_per_start, dtype=float)
    # ... compute rotations of subunits in sub-helix
    subunit_rotations_in_subhelix = compute_subunit_rotations_in_subhelix(
        subunit_indices, twist, initial_rotation
    )

    # Now, transform single sub-helix rotations into all sub-helices rotations,
    # related by C_n symmetry
    def compute_helix_subunit_rotations_per_start(symmetry_angle, R):
        # ... rotate the sub-helix around the screw axis to a different sub-helix
        c_n, s_n = jnp.cos(symmetry_angle), jnp.sin(symmetry_angle)
        R_n = jnp.array(((c_n, -s_n, 0), (s_n, c_n, 0), (0, 0, 1)), dtype=float)
        return jnp.einsum("ij,njk->nik", R_n, R)

    # ... function to rotate subunit about symmetry angle
    compute_helix_subunit_rotations = jax.vmap(
        compute_helix_subunit_rotations_per_start, in_axes=[0, None]
    )
    # ... compute symmetry angles relating first sub-helix to all other sub-helices
    symmetry_angles = jnp.array([2 * jnp.pi * n / n_start for n in range(n_start)])
    # ... finally, get all subunit rotations!
    subunit_rotations = compute_helix_subunit_rotations(
        symmetry_angles, subunit_rotations_in_subhelix
    ).reshape((n_start * n_subunits_per_start, 3, 3))

    return subunit_rotations



"""
Abstraction of the ice in a cryo-EM image.
"""

from abc import abstractmethod
from typing_extensions import override

import jax.numpy as jnp
import jax.random as jr
import numpy as np
from equinox import Module
from jaxtyping import Array, Complex, PRNGKeyArray

from ..image.operators import FourierOperatorLike
from ._config import ImageConfig


class AbstractIce(Module, strict=True):
    """Base class for an ice model."""

    @abstractmethod
    def sample(
        self, key: PRNGKeyArray, config: ImageConfig
    ) -> Complex[Array, "{config.padded_y_dim} {config.padded_x_dim//2+1}"]:
        """Sample a stochastic realization of the phase shifts due to the ice
        at the exit plane."""
        raise NotImplementedError

    def __call__(
        self,
        key: PRNGKeyArray,
        fourier_phase_at_exit_plane: Complex[
            Array, "{config.padded_y_dim} {config.padded_x_dim//2+1}"
        ],
        config: ImageConfig,
    ) -> Complex[Array, "{config.padded_y_dim} {config.padded_x_dim//2+1}"]:
        """Compute the combined phase of the ice and the specimen."""
        # Sample the realization of the phase due to the ice.
        fourier_ice_phase_at_exit_plane = self.sample(key, config)

        return fourier_phase_at_exit_plane + fourier_ice_phase_at_exit_plane


class GaussianIce(AbstractIce, strict=True):
    r"""Ice modeled as gaussian noise.

    **Attributes:**

    - `variance` : A function that computes the variance
                 of the ice, modeled as colored gaussian noise.
                 The dimensions of this function are a squared
                 phase contrast.
    """

    variance: FourierOperatorLike

    def __init__(self, variance: FourierOperatorLike):
        self.variance = variance

    @override
    def sample(
        self, key: PRNGKeyArray, config: ImageConfig
    ) -> Complex[Array, "{config.padded_y_dim} {config.padded_x_dim//2+1}"]:
        """Sample a realization of the ice phase shifts as colored gaussian noise."""
        N_pix = np.prod(config.padded_shape)
        frequency_grid_in_angstroms = (
            config.wrapped_padded_frequency_grid_in_angstroms.get()
        )
        # Compute standard deviation, scaling up by the variance by the number
        # of pixels to make the realization independent pixel-independent in real-space.
        std = jnp.sqrt(N_pix * self.variance(frequency_grid_in_angstroms))
        ice_phase_at_exit_plane = std * jr.normal(
            key,
            shape=frequency_grid_in_angstroms.shape[0:-1],
            dtype=complex,
        ).at[0, 0].set(0.0)

        return ice_phase_at_exit_plane



from ._dataset import AbstractDataset as AbstractDataset
from ._particle_stack import (
    AbstractParticleStack as AbstractParticleStack,
)
from ._relion import (
    default_relion_make_config as default_relion_make_config,
    HelicalRelionDataset as HelicalRelionDataset,
    RelionDataset as RelionDataset,
    RelionParticleStack as RelionParticleStack,
)



from ._assembly import (
    AbstractAssembly as AbstractAssembly,
    compute_helical_lattice_positions as compute_helical_lattice_positions,
    compute_helical_lattice_rotations as compute_helical_lattice_rotations,
    Helix as Helix,
)
from ._config import ImageConfig as ImageConfig
from ._conformation import (
    AbstractConformation as AbstractConformation,
    DiscreteConformation as DiscreteConformation,
)
from ._detector import (
    AbstractDetector as AbstractDetector,
    AbstractDQE as AbstractDQE,
    GaussianDetector as GaussianDetector,
    IdealDQE as IdealDQE,
    PoissonDetector as PoissonDetector,
)
from ._dose import ElectronDose as ElectronDose
from ._ice import (
    AbstractIce as AbstractIce,
    GaussianIce as GaussianIce,
)
from ._instrument import Instrument as Instrument
from ._integrators import (
    AbstractPotentialIntegrator as AbstractPotentialIntegrator,
    extract_slice as extract_slice,
    extract_slice_with_cubic_spline as extract_slice_with_cubic_spline,
    FourierSliceExtract as FourierSliceExtract,
    NufftProject as NufftProject,
    project_with_nufft as project_with_nufft,
)
from ._optics import (
    AbstractOptics as AbstractOptics,
    CTF as CTF,
    WeakPhaseOptics as WeakPhaseOptics,
)
from ._pipeline import (
    AbstractPipeline as AbstractPipeline,
    AssemblyPipeline as AssemblyPipeline,
    ImagePipeline as ImagePipeline,
)
from ._pose import (
    AbstractPose as AbstractPose,
    AxisAnglePose as AxisAnglePose,
    EulerAnglePose as EulerAnglePose,
    QuaternionPose as QuaternionPose,
)
from ._potential import (
    AbstractFourierVoxelGridPotential as AbstractFourierVoxelGridPotential,
    AbstractScatteringPotential as AbstractScatteringPotential,
    AbstractVoxelPotential as AbstractVoxelPotential,
    build_real_space_voxels_from_atoms as build_real_space_voxels_from_atoms,
    evaluate_3d_atom_potential as evaluate_3d_atom_potential,
    evaluate_3d_real_space_gaussian as evaluate_3d_real_space_gaussian,
    FourierVoxelGridPotential as FourierVoxelGridPotential,
    FourierVoxelGridPotentialInterpolator as FourierVoxelGridPotentialInterpolator,
    RealVoxelCloudPotential as RealVoxelCloudPotential,
    RealVoxelGridPotential as RealVoxelGridPotential,
)
from ._specimen import (
    AbstractEnsemble as AbstractEnsemble,
    AbstractSpecimen as AbstractSpecimen,
    DiscreteEnsemble as DiscreteEnsemble,
    Specimen as Specimen,
)



import os

import equinox as eqx
import jax
import jax.numpy as jnp
import jax.random as jr
import pytest
from jaxtyping import install_import_hook


with install_import_hook("cryojax", "typeguard.typechecked"):
    import cryojax as cryojax
    import cryojax.simulator as cs
    from cryojax.image import operators as op, rfftn
    from cryojax.io import read_array_with_spacing_from_mrc


# jax.config.update("jax_numpy_dtype_promotion", "strict")
# jax.config.update("jax_numpy_rank_promotion", "raise")
jax.config.update("jax_enable_x64", True)


@pytest.fixture
def sample_mrc_path():
    return os.path.join(os.path.dirname(__file__), "data", "3j9g_potential_ps4_4.mrc")


@pytest.fixture
def sample_subunit_mrc_path():
    return os.path.join(
        os.path.dirname(__file__), "data", "3j9g_subunit_potential_ps4_4.mrc"
    )


@pytest.fixture
def sample_pdb_path():
    return os.path.join(os.path.dirname(__file__), "data", "1uao.pdb")


@pytest.fixture
def toy_gaussian_cloud():
    atom_positions = jnp.array(
        [
            [0.0, 0.0, 0.0],
            [0.0, 0.0, 1.0],
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
        ]
    )
    num_atoms = atom_positions.shape[0]
    ff_a = jnp.array(
        num_atoms
        * [
            [1.0, 0.5],
        ]
    )

    ff_b = jnp.array(
        num_atoms
        * [
            [0.3, 0.2],
        ]
    )

    n_voxels_per_side = (128, 128, 128)
    voxel_size = 0.05
    return (atom_positions, ff_a, ff_b, n_voxels_per_side, voxel_size)


@pytest.fixture
def pixel_size():
    return 4.4


@pytest.fixture
def config(pixel_size):
    return cs.ImageConfig((65, 66), pixel_size, pad_scale=1.1)


@pytest.fixture
def integrator():
    return cs.FourierSliceExtract(interpolation_order=1)


@pytest.fixture
def potential(sample_mrc_path):
    real_voxel_grid, voxel_size = read_array_with_spacing_from_mrc(sample_mrc_path)
    return cs.FourierVoxelGridPotential.from_real_voxel_grid(
        real_voxel_grid, voxel_size, pad_scale=1.3
    )


@pytest.fixture
def filters(config):
    return op.LowpassFilter(config.wrapped_padded_frequency_grid_in_pixels.get())


@pytest.fixture
def masks(config):
    return op.CircularMask(
        config.wrapped_padded_coordinate_grid_in_angstroms.get(),
        radius=20 * float(config.pixel_size),
    )


@pytest.fixture
def instrument():
    voltage_in_kilovolts = 300.0
    return cs.Instrument(
        voltage_in_kilovolts,
        optics=cs.WeakPhaseOptics(cs.CTF()),
        dose=cs.ElectronDose(electrons_per_angstrom_squared=1000.0),
        detector=cs.GaussianDetector(cs.IdealDQE(fraction_detected_electrons=1.0)),
    )


@pytest.fixture
def pose():
    return cs.EulerAnglePose(
        view_phi=30.0,
        view_theta=100.0,
        view_psi=-10.0,
        offset_x_in_angstroms=10.0,
        offset_y_in_angstroms=-5.0,
    )


@pytest.fixture
def specimen(potential, integrator, pose):
    return cs.Specimen(potential, integrator, pose)


@pytest.fixture
def solvent():
    return cs.GaussianIce(op.Constant(0.001**2))


@pytest.fixture
def noiseless_model(config, specimen, instrument):
    instrument = eqx.tree_at(lambda ins: ins.detector, instrument, None)
    return cs.ImagePipeline(config=config, specimen=specimen, instrument=instrument)


@pytest.fixture
def noisy_model(config, specimen, instrument, solvent):
    return cs.ImagePipeline(
        config=config,
        specimen=specimen,
        instrument=instrument,
        solvent=solvent,
    )


@pytest.fixture
def filtered_model(config, specimen, instrument, solvent, filters):
    return cs.ImagePipeline(
        config=config,
        specimen=specimen,
        instrument=instrument,
        solvent=solvent,
        filter=filters,
    )


@pytest.fixture
def filtered_and_masked_model(config, specimen, instrument, solvent, filters, masks):
    return cs.ImagePipeline(
        config=config,
        specimen=specimen,
        instrument=instrument,
        solvent=solvent,
        filter=filters,
        mask=masks,
    )


@pytest.fixture
def test_image(noisy_model):
    image = noisy_model.sample(jr.PRNGKey(1234))
    return rfftn(image)



"""
Coordinate functionality in cryojax.
"""

from abc import abstractmethod
from typing import Any, Optional
from typing_extensions import Self

import equinox as eqx
import jax.numpy as jnp
import numpy as np
from equinox import AbstractVar
from jaxtyping import Array, Float, Inexact


class AbstractCoordinates(eqx.Module, strict=True):
    """
    A base class that wraps a coordinate array.
    """

    array: AbstractVar[Any]

    @abstractmethod
    def get(self) -> Any:
        """Get the coordinates."""
        raise NotImplementedError

    def __mul__(
        self, real_number: float | Float[np.ndarray, ""] | Float[Array, ""]
    ) -> Self:
        # The following line seems to be required for differentiability with
        # respect to arr
        rescaled_array = jnp.where(
            self.array != 0.0, self.array * jnp.asarray(real_number), 0.0
        )
        return eqx.tree_at(lambda x: x.array, self, rescaled_array)

    def __rmul__(
        self, real_number: float | Float[np.ndarray, ""] | Float[Array, ""]
    ) -> Self:
        rescaled_array = jnp.where(
            self.array != 0.0, jnp.asarray(real_number) * self.array, 0.0
        )
        return eqx.tree_at(lambda x: x.array, self, rescaled_array)

    def __truediv__(
        self, real_number: float | Float[np.ndarray, ""] | Float[Array, ""]
    ) -> Self:
        rescaled_array = jnp.where(
            self.array != 0.0, self.array / jnp.asarray(real_number), 0.0
        )
        return eqx.tree_at(lambda x: x.array, self, rescaled_array)


class CoordinateList(AbstractCoordinates, strict=True):
    """
    A Pytree that wraps a coordinate list.
    """

    array: Float[Array, "size 3"] | Float[Array, "size 2"] = eqx.field(
        converter=jnp.asarray
    )

    def __init__(
        self, coordinate_list: Float[Array, "size 2"] | Float[Array, "size 3"]
    ):
        self.array = coordinate_list

    def get(self) -> Float[Array, "size 3"] | Float[Array, "size 2"]:
        return self.array


class CoordinateGrid(AbstractCoordinates, strict=True):
    """
    A Pytree that wraps a coordinate grid.
    """

    array: Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"] = (
        eqx.field(converter=jnp.asarray)
    )

    def __init__(
        self,
        shape: tuple[int, ...],
        grid_spacing: float | Float[np.ndarray, ""] = 1.0,
    ):
        self.array = make_coordinates(shape, grid_spacing)

    def get(
        self,
    ) -> Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]:
        return self.array


class FrequencyGrid(AbstractCoordinates, strict=True):
    """
    A Pytree that wraps a frequency grid.
    """

    array: Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"] = (
        eqx.field(converter=jnp.asarray)
    )

    def __init__(
        self,
        shape: tuple[int, ...],
        grid_spacing: float | Float[np.ndarray, ""] = 1.0,
        half_space: bool = True,
    ):
        self.array = make_frequencies(shape, grid_spacing, half_space=half_space)

    def get(
        self,
    ) -> Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"]:
        return self.array


class FrequencySlice(AbstractCoordinates, strict=True):
    """
    A Pytree that wraps a frequency slice.

    Unlike a `FrequencyGrid`, a `FrequencySlice` has the zero frequency
    component in the center.
    """

    array: Float[Array, "1 y_dim x_dim 3"] = eqx.field(converter=jnp.asarray)

    def __init__(
        self,
        shape: tuple[int, int],
        grid_spacing: float | Float[np.ndarray, ""] = 1.0,
        half_space: bool = True,
    ):
        frequency_slice = make_frequencies(shape, grid_spacing, half_space=half_space)
        if half_space:
            frequency_slice = jnp.fft.fftshift(frequency_slice, axes=(0,))
        else:
            frequency_slice = jnp.fft.fftshift(frequency_slice, axes=(0, 1))
        frequency_slice = jnp.expand_dims(
            jnp.pad(
                frequency_slice,
                ((0, 0), (0, 0), (0, 1)),
                mode="constant",
                constant_values=0.0,
            ),
            axis=0,
        )
        self.array = frequency_slice

    def get(self) -> Float[Array, "1 y_dim x_dim 3"]:
        return self.array


def make_coordinates(
    shape: tuple[int, ...], grid_spacing: float | Float[np.ndarray, ""] = 1.0
) -> Float[Array, "*shape ndim"]:
    """
    Create a real-space cartesian coordinate system on a grid.

    Arguments
    ---------
    shape :
        Shape of the voxel grid, with
        ``ndim = len(shape)``.
    grid_spacing :
        The grid spacing, in units of length.

    Returns
    -------
    coordinate_grid :
        Cartesian coordinate system in real space.
    """
    coordinate_grid = _make_coordinates_or_frequencies(
        shape, grid_spacing=grid_spacing, real_space=True
    )
    return coordinate_grid


def make_frequencies(
    shape: tuple[int, ...],
    grid_spacing: float | Float[np.ndarray, ""] = 1.0,
    half_space: bool = True,
) -> Float[Array, "*shape ndim"]:
    """
    Create a fourier-space cartesian coordinate system on a grid.
    The zero-frequency component is in the beginning.

    Arguments
    ---------
    shape :
        Shape of the voxel grid, with
        ``ndim = len(shape)``.
    grid_spacing :
        The grid spacing, in units of length.
    half_space :
        Return a frequency grid on the half space.
        ``shape[-1]`` is the axis on which the negative
        frequencies are omitted.

    Returns
    -------
    frequency_grid :
        Cartesian coordinate system in frequency space.
    """
    frequency_grid = _make_coordinates_or_frequencies(
        shape,
        grid_spacing=grid_spacing,
        real_space=False,
        half_space=half_space,
    )
    return frequency_grid


def cartesian_to_polar(
    freqs: Float[Array, "y_dim x_dim 2"], square: bool = False
) -> tuple[Inexact[Array, "y_dim x_dim"], Inexact[Array, "y_dim x_dim"]]:
    """
    Convert from cartesian to polar coordinates.

    Arguments
    ---------
    freqs :
        The cartesian coordinate system.
    square :
        If ``True``, return the square of the
        radial coordinate :math:`|r|^2`. Otherwise,
        return :math:`|r|`.
    """
    theta = jnp.arctan2(freqs[..., 0], freqs[..., 1])
    k_sqr = jnp.sum(jnp.square(freqs), axis=-1)
    if square:
        return k_sqr, theta
    else:
        kr = jnp.sqrt(k_sqr)
        return kr, theta


def _make_coordinates_or_frequencies(
    shape: tuple[int, ...],
    grid_spacing: float | Float[np.ndarray, ""] = 1.0,
    real_space: bool = False,
    half_space: bool = True,
) -> Float[Array, "*shape ndim"]:
    ndim = len(shape)
    coords1D = []
    for idx in range(ndim):
        if real_space:
            c1D = _make_coordinates_or_frequencies_1d(
                shape[idx], grid_spacing, real_space
            )
        else:
            if not half_space:
                rfftfreq = False
            else:
                rfftfreq = False if idx < ndim - 1 else True
            c1D = _make_coordinates_or_frequencies_1d(
                shape[idx], grid_spacing, real_space, rfftfreq
            )
        coords1D.append(c1D)
    if ndim == 2:
        y, x = coords1D
        xv, yv = jnp.meshgrid(x, y, indexing="xy")
        coords = jnp.stack([xv, yv], axis=-1)
    elif ndim == 3:
        z, y, x = coords1D
        xv, yv, zv = jnp.meshgrid(x, y, z, indexing="xy")
        xv, yv, zv = [
            jnp.transpose(rv, axes=[2, 0, 1]) for rv in [xv, yv, zv]
        ]  # Change axis ordering to [z, y, x]
        coords = jnp.stack([xv, yv, zv], axis=-1)
    else:
        raise ValueError(
            "Only 2D and 3D coordinate grids are supported. "
            f"Tried to create a grid of shape {shape}."
        )

    return coords


def _make_coordinates_or_frequencies_1d(
    size: int,
    grid_spacing: float | Float[np.ndarray, ""],
    real_space: bool = False,
    rfftfreq: Optional[bool] = None,
) -> Float[Array, " size"]:
    """One-dimensional coordinates in real or fourier space"""
    if real_space:
        make_1d = (
            lambda size, dx: jnp.fft.fftshift(jnp.fft.fftfreq(size, 1 / dx)) * size
        )
    else:
        if rfftfreq is None:
            raise ValueError("Argument rfftfreq cannot be None if real_space=False.")
        else:
            fn = jnp.fft.rfftfreq if rfftfreq else jnp.fft.fftfreq
            make_1d = lambda size, dx: fn(size, grid_spacing)

    return make_1d(size, grid_spacing)



import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np
import pytest

import cryojax.simulator as cs
from cryojax.io import read_array_with_spacing_from_mrc


def build_helix(sample_subunit_mrc_path, n_subunits_per_start) -> cs.Helix:
    real_voxel_grid, voxel_size = read_array_with_spacing_from_mrc(
        sample_subunit_mrc_path
    )
    subunit_density = cs.FourierVoxelGridPotential.from_real_voxel_grid(
        real_voxel_grid, voxel_size, pad_scale=2
    )
    integrator = cs.FourierSliceExtract()
    r_0 = jnp.asarray([-88.70895129, 9.75357114, 0.0], dtype=float)
    subunit_pose = cs.EulerAnglePose(*r_0)
    subunit = cs.Specimen(subunit_density, integrator, subunit_pose)
    return cs.Helix(
        subunit,
        rise=21.8,
        twist=29.4,
        n_start=6,
        n_subunits=n_subunits_per_start * 6,
    )


def build_helix_with_conformation(
    sample_subunit_mrc_path, n_subunits_per_start
) -> cs.Helix:
    subunit_density = tuple(
        [
            cs.FourierVoxelGridPotential.from_real_voxel_grid(
                *read_array_with_spacing_from_mrc(sample_subunit_mrc_path)
            )
            for _ in range(2)
        ]
    )
    n_start = 6
    r_0 = jnp.asarray([-88.70895129, 9.75357114, 0.0], dtype=float)
    subunit_pose = cs.EulerAnglePose(*r_0)
    integrator = cs.FourierSliceExtract()
    subunit = cs.DiscreteEnsemble(
        subunit_density,
        integrator,
        subunit_pose,
        conformation=cs.DiscreteConformation(0),
    )
    conformation = jax.vmap(lambda value: cs.DiscreteConformation(value))(
        np.random.choice(2, n_start * n_subunits_per_start)
    )
    return cs.Helix(
        subunit,
        conformation=conformation,
        rise=21.8,
        twist=29.4,
        n_start=n_start,
        n_subunits=n_subunits_per_start * 6,
    )


def test_superposition_pipeline_without_conformation(sample_subunit_mrc_path, config):
    helix = build_helix(sample_subunit_mrc_path, 1)
    pipeline = cs.AssemblyPipeline(
        config=config, assembly=helix, instrument=cs.Instrument(300.0)
    )
    _ = pipeline.render()
    _ = pipeline.sample(jax.random.PRNGKey(0))


def test_superposition_pipeline_with_conformation(sample_subunit_mrc_path, config):
    helix = build_helix_with_conformation(sample_subunit_mrc_path, 2)
    pipeline = cs.AssemblyPipeline(
        config=config, instrument=cs.Instrument(300.0), assembly=helix
    )
    _ = pipeline.render()
    _ = pipeline.sample(jax.random.PRNGKey(0))


@pytest.mark.parametrize(
    "rotation_angle, n_subunits_per_start",
    [(360.0 / 6, 1), (2 * 360.0 / 6, 1), (360.0 / 6, 2)],
)
def test_c6_rotation(
    sample_subunit_mrc_path, config, rotation_angle, n_subunits_per_start
):
    helix = build_helix(sample_subunit_mrc_path, n_subunits_per_start)

    @jax.jit
    def compute_rotated_image(config, helix, pose):
        helix = eqx.tree_at(lambda m: m.pose, helix, pose)
        pipeline = cs.AssemblyPipeline(
            config=config, instrument=cs.Instrument(300.0), assembly=helix
        )
        return pipeline.render(normalize=True)

    np.testing.assert_allclose(
        compute_rotated_image(config, helix, cs.EulerAnglePose()),
        compute_rotated_image(
            config, helix, cs.EulerAnglePose(view_phi=rotation_angle)
        ),
    )


@pytest.mark.parametrize(
    "translation, euler_angles",
    [
        ((0.0, 0.0), (60.0, 100.0, -40.0)),
        ((1.0, -3.0), (10.0, 50.0, 100.0)),
    ],
)
def test_agree_with_3j9g_assembly(
    sample_subunit_mrc_path, potential, config, translation, euler_angles
):
    instrument = cs.Instrument(voltage_in_kilovolts=300.0)
    helix = build_helix(sample_subunit_mrc_path, 2)
    specimen_39jg = cs.Specimen(potential, helix.subunit.integrator)
    pipeline_for_assembly = cs.AssemblyPipeline(
        config=config, instrument=instrument, assembly=helix
    )
    pipeline_for_3j9g = cs.ImagePipeline(
        config=config, instrument=instrument, specimen=specimen_39jg
    )

    @eqx.filter_jit
    def compute_rotated_image_with_helix(
        pipeline: cs.AssemblyPipeline, pose: cs.AbstractPose
    ):
        pipeline = eqx.tree_at(lambda m: m.assembly.pose, pipeline, pose)
        return pipeline.render(normalize=True)

    @eqx.filter_jit
    def compute_rotated_image_with_3j9g(
        pipeline: cs.ImagePipeline, pose: cs.AbstractPose
    ):
        pipeline = eqx.tree_at(lambda m: m.specimen.pose, pipeline, pose)
        return pipeline.render(normalize=True)

    pose = cs.EulerAnglePose(*translation, 0.0, *euler_angles)
    reference_image = compute_rotated_image_with_3j9g(
        pipeline_for_3j9g, cs.EulerAnglePose()
    )
    assembled_image = compute_rotated_image_with_helix(pipeline_for_assembly, pose)
    test_image = compute_rotated_image_with_3j9g(pipeline_for_3j9g, pose)
    assert np.std(assembled_image - test_image) < 10 * np.std(
        assembled_image - reference_image
    )


def test_transform_by_rise_and_twist(sample_subunit_mrc_path, pixel_size):
    helix = build_helix(sample_subunit_mrc_path, 12)
    config = cs.ImageConfig((50, 20), pixel_size, pad_scale=6)

    @jax.jit
    def compute_rotated_image(config, helix, pose):
        helix = eqx.tree_at(lambda m: m.pose, helix, pose)
        pipeline = cs.AssemblyPipeline(
            config=config, instrument=cs.Instrument(300.0), assembly=helix
        )
        return pipeline.render(normalize=True)

    np.testing.assert_allclose(
        compute_rotated_image(
            config,
            helix,
            cs.EulerAnglePose(view_phi=0.0, view_theta=90.0, view_psi=0.0),
        ),
        compute_rotated_image(
            config,
            helix,
            cs.EulerAnglePose(
                view_phi=helix.twist,
                view_theta=90.0,
                view_psi=0.0,
                offset_x_in_angstroms=helix.rise,
            ),
        ),
        atol=1e-1,
    )



"""
Transformations used for reparameterizing cryojax models.
"""

from abc import abstractmethod
from typing import Any, Callable, Optional, Sequence, Union
from typing_extensions import overload

import equinox as eqx
import jax
import jax.numpy as jnp
import jax.tree_util as jtu
from equinox import AbstractVar, field, Module
from jaxtyping import Array, Float, PyTree

from ...core import error_if_not_positive, error_if_zero


def _is_transformed(x: Any) -> bool:
    return isinstance(x, AbstractParameterTransform)


def _resolve_transform(x: Any) -> Any:
    if isinstance(x, AbstractParameterTransform):
        return x.get()
    else:
        return x


def _apply_transform(
    pytree: PyTree,
    where: Callable[[PyTree], Union[Any, Sequence[Any]]],
    replace_fn: Callable[[Any], "AbstractParameterTransform"],
    is_leaf: Optional[Callable[[Any], bool]] = None,
) -> PyTree:
    return eqx.tree_at(where, pytree, replace_fn=replace_fn, is_leaf=is_leaf)


@overload
def insert_transforms(
    pytree: PyTree,
    wheres: Sequence[Callable[[PyTree], Union[Any, Sequence[Any]]]],
    replace_fns: Sequence[Callable[[Any], "AbstractParameterTransform"]],
    *,
    is_leaf: Optional[Callable[[Any], bool]] = None,
) -> PyTree: ...


@overload
def insert_transforms(
    pytree: PyTree,
    wheres: Callable[[PyTree], Union[Any, Sequence[Any]]],
    replace_fns: Callable[[Any], "AbstractParameterTransform"],
    *,
    is_leaf: Optional[Callable[[Any], bool]] = None,
) -> PyTree: ...


def insert_transforms(
    pytree: PyTree,
    wheres: (
        Callable[[PyTree], Union[Any, Sequence[Any]]]
        | Sequence[Callable[[PyTree], Union[Any, Sequence[Any]]]]
    ),
    replace_fns: (
        Callable[[Any], "AbstractParameterTransform"]
        | Sequence[Callable[[Any], "AbstractParameterTransform"]]
    ),
    *,
    is_leaf: Optional[Callable[[Any], bool]] = None,
) -> PyTree:
    """Applies an `AbstractParameterTransform` to pytree node(s).

    This function performs a sequence of `equinox.tree_at` calls to apply each
    `replace_fn` in `replace_fns` to each `where` in `wheres`.
    """
    if isinstance(replace_fns, Callable) and isinstance(wheres, Callable):
        where, replace_fn = wheres, replace_fns
        return _apply_transform(pytree, where, replace_fn, is_leaf=is_leaf)
    elif isinstance(replace_fns, Sequence) and isinstance(wheres, Sequence):
        if len(replace_fns) != len(wheres):
            raise TypeError(
                "If arguments `wheres` and `replace_fns` are sequences, they "
                "must be sequences of the same length. Got "
                f"`wheres, replace_fns = {wheres}, {replace_fns}`."
            )
        transformed_pytree = pytree
        for where, replace_fn in zip(wheres, replace_fns):
            transformed_pytree = _apply_transform(
                pytree, where, replace_fn, is_leaf=is_leaf
            )
        return transformed_pytree
    else:
        raise TypeError(
            "Input arguments `wheres` and `replace_fns` must both either be functions "
            f"or sequences. Got `wheres, replace_fns = {wheres}, {replace_fns}`."
        )


def resolve_transforms(pytree: PyTree) -> PyTree:
    """Transforms a pytree whose parameters have entries
    that are `AbstractParameterTransform`s back to its
    original parameter space.
    """
    return jtu.tree_map(_resolve_transform, pytree, is_leaf=_is_transformed)


class AbstractParameterTransform(Module, strict=True):
    """Base class for a parameter transformation.

    This interface tries to implement a user-friendly way to get custom
    per-parameter behavior in pytrees, as described in the [equinox docs](
    https://docs.kidger.site/equinox/tricks/#custom-per-parameter-behaviour).

    Typically this is used to do inference in better geometries or to enforce
    parameter constraints, however it can be used generally.

    This is a very simple class. When the class is initialized,
    a parameter is taken to a transformed parameter space. When
    `transform.get()` is called, the parameter is taken back to
    the original parameter space.
    """

    transformed_parameter: AbstractVar[Array]

    @abstractmethod
    def get(self) -> Any:
        """Get the parameter in the original parameter space."""
        raise NotImplementedError


class ExpTransform(AbstractParameterTransform, strict=True):
    """This class transforms a parameter to its logarithm.

    **Attributes:**

    - `transformed_parameter`: The parameter on a logarithmic scale.
    """

    transformed_parameter: Array

    def __init__(self, parameter: Array):
        """**Arguments:**

        - `parameter`: The parameter to be transformed to a logarithmic
                       scale.
        """
        self.transformed_parameter = jnp.log(error_if_not_positive(parameter))

    def get(self) -> Array:
        """The logarithmic-valued parameter transformed with an exponential."""
        return jnp.exp(self.transformed_parameter)


class RescalingTransform(AbstractParameterTransform, strict=True):
    """This class transforms a parameter by a scale factor and a shift.

    **Attributes:**

    - `transformed_parameter`: The rescaled parameter.
    """

    transformed_parameter: Array
    scaling: Float[Array, ""] = field(converter=error_if_zero)
    shift: Float[Array, ""]

    def __init__(
        self,
        parameter: Array,
        scaling: Float[Array, ""] | float,
        shift: Float[Array, ""] | float = 0.0,
    ):
        """**Arguments:**

        - `parameter`: The parameter to be rescaled.

        - `scaling`: The scale factor.

        - `shift`: The shift.
        """
        self.scaling = jnp.asarray(scaling)
        self.shift = jnp.asarray(shift)
        self.transformed_parameter = self.scaling * jnp.asarray(parameter) + self.shift

    def get(self) -> Array:
        """The rescaled parameter transformed back to the original scale."""
        return (
            self.transformed_parameter - jax.lax.stop_gradient(self.shift)
        ) / jax.lax.stop_gradient(self.scaling)


class ComposedTransform(AbstractParameterTransform, strict=True):
    """This class composes multiple `AbstractParameterTransforms`.

    **Attributes:**

    - `transformed_parameter`: The transformed parameter.

    - `transforms`: The sequence of `AbstractParameterTransform`s.
    """

    transformed_parameter: Any
    transforms: tuple[AbstractParameterTransform, ...]

    def __init__(
        self,
        parameter: Any,
        transform_fns: Sequence[Callable[[Any], "AbstractParameterTransform"]],
    ):
        """**Arguments:**

        - `parameter`: The parameter to be transformed.

        - `transform_fns`: A sequence of functions that take in
                           a parameter and return an `AbstractParameterTransform`.
        """
        p = jnp.asarray(parameter)
        transforms = []
        for transform_fn in transform_fns:
            transform = transform_fn(p)
            p = transform.transformed_parameter
            transforms.append(transform)
        self.transformed_parameter = p
        self.transforms = tuple(transforms)

    def get(self) -> Any:
        """Transform the `transformed_parameter` back to the original space."""
        parameter = self.transformed_parameter
        transforms = jax.lax.stop_gradient(self.transforms)
        for transform in transforms[::-1]:
            transform = eqx.tree_at(
                lambda t: t.transformed_parameter, transform, parameter
            )
            parameter = transform.get()
        return parameter



"""
Abstraction of a biological assembly. This assembles a structure
by computing an Ensemble of subunits, parameterized by
some geometry.
"""

from abc import abstractmethod
from functools import cached_property
from typing import Optional

import equinox as eqx
import jax
from equinox import AbstractVar
from jaxtyping import Array, Float

from ...rotations import SO3
from .._conformation import AbstractConformation
from .._pose import AbstractPose
from .._specimen import AbstractEnsemble, AbstractSpecimen


class AbstractAssembly(eqx.Module, strict=True):
    """Abstraction of a biological assembly.

    This class acts just like an ``AbstractSpecimen``, however
    it creates an assembly from a subunit.

    To subclass an `AbstractAssembly`,
        1) Overwrite the `AbstractAssembly.n_subunits`
           property
        2) Overwrite the `AbstractAssembly.positions`
           and `AbstractAssembly.rotations` properties.
    """

    subunit: AbstractVar[AbstractSpecimen]
    pose: AbstractVar[AbstractPose]
    conformation: AbstractVar[Optional[AbstractConformation]]

    n_subunits: AbstractVar[int]

    def __check_init__(self):
        if self.conformation is not None and not isinstance(
            self.subunit, AbstractEnsemble
        ):
            # Make sure that if conformation is set, subunit is an AbstractEnsemble
            raise AttributeError(
                f"If {type(self)}.conformation is set, {type(self)}.subunit must be an "
                "AbstractEnsemble."
            )
        if self.conformation is not None and isinstance(self.subunit, AbstractEnsemble):
            # ... if it is an AbstractEnsemble, the AbstractConformation must be the
            #  right type
            if not isinstance(self.conformation, type(self.subunit.conformation)):
                raise AttributeError(
                    f"{type(self)}.conformation must be type "
                    f" {type(self.subunit.conformation)} if {type(self)}.subunit is "
                    f"type {type(self.subunit)}."
                )

    @cached_property
    @abstractmethod
    def offsets_in_angstroms(self) -> Float[Array, "{n_subunits} 3"]:
        """The positions of each subunit."""
        raise NotImplementedError

    @cached_property
    @abstractmethod
    def rotations(self) -> SO3:
        """The relative rotations between subunits."""
        raise NotImplementedError

    @cached_property
    def poses(self) -> AbstractPose:
        """
        Draw the poses of the subunits in the lab frame, measured
        from the rotation relative to the first subunit.
        """
        # Transform the subunit positions by pose of the helix
        transformed_positions = (
            self.pose.rotate_coordinates(self.offsets_in_angstroms, inverse=False)
            + self.pose.offset_in_angstroms
        )
        # Transform the subunit rotations by the pose of the helix. This operation
        # left multiplies by the pose of the helix, taking care that first subunits
        # are rotated to the center of mass frame, then the lab frame.
        transformed_rotations = jax.vmap(
            lambda com_rotation, subunit_rotation: com_rotation @ subunit_rotation,
            in_axes=[None, 0],
        )(self.pose.rotation, self.rotations)
        # Function to construct AbstractPoses
        cls = type(self.pose)
        make_assembly_poses = jax.vmap(
            lambda rot, pos: cls.from_rotation_and_translation(rot, pos)
        )

        return make_assembly_poses(transformed_rotations, transformed_positions)

    @cached_property
    def subunits(self) -> AbstractSpecimen:
        """Draw a realization of all of the subunits in the lab frame."""
        # Compute a list of subunits, configured at the correct conformations
        if isinstance(self.subunit, AbstractEnsemble):
            where = lambda s: (s.conformation, s.pose)
            return eqx.tree_at(where, self.subunit, (self.conformation, self.poses))
        else:
            where = lambda s: s.pose
            return eqx.tree_at(where, self.subunit, self.poses)



"""
Routines to compute FFTs, in cryojax conventions.
"""

from typing import Any, Optional, overload

import jax.numpy as jnp
from jaxtyping import Array, Complex, Float, Inexact


@overload
def ifftn(
    ft: Complex[Array, "z_dim y_dim x_dim"],
    real: bool = False,
    axes: Optional[tuple[int, ...]] = None,
    **kwargs: Any,
) -> Complex[Array, "z_dim y_dim x_dim"]: ...


@overload
def ifftn(
    ft: Complex[Array, "y_dim x_dim"],
    real: bool = False,
    axes: Optional[tuple[int, ...]] = None,
    **kwargs: Any,
) -> Complex[Array, "y_dim x_dim"]: ...


def ifftn(
    ft: Complex[Array, "y_dim x_dim"] | Complex[Array, "z_dim y_dim x_dim"],
    real: bool = False,
    axes: Optional[tuple[int, ...]] = None,
    **kwargs: Any,
) -> Complex[Array, "y_dim x_dim"] | Complex[Array, "z_dim y_dim x_dim"]:
    """
    Helper routine to match the inverse fourier transform
    to the output of a type 1 non-uniform FFT with jax-finufft.

    Arguments
    ---------
    ft :
        Fourier transform array. Assumes that the zero
        frequency component is in the center.
    real : `bool`
        If ``True``, then ``ft`` is the Fourier transform
        of a real-valued function.
    Returns
    -------
    ift :
        Inverse fourier transform.
    """
    ift = jnp.fft.fftshift(jnp.fft.ifftn(ft, axes=axes, **kwargs), axes=axes)

    if real:
        return ift.real
    else:
        return ift


@overload
def fftn(
    ift: Inexact[Array, "z_dim y_dim x_dim"],
    axes: Optional[tuple[int, ...]] = None,
    **kwargs: Any,
) -> Complex[Array, "z_dim y_dim x_dim"]: ...


@overload
def fftn(
    ift: Inexact[Array, "y_dim x_dim"],
    axes: Optional[tuple[int, ...]] = None,
    **kwargs: Any,
) -> Complex[Array, "y_dim x_dim"]: ...


def fftn(
    ift: Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"],
    axes: Optional[tuple[int, ...]] = None,
    **kwargs: Any,
) -> Complex[Array, "y_dim x_dim"] | Complex[Array, "z_dim y_dim x_dim"]:
    """
    Helper routine to match the fourier transform of an array
    with the output of a type 1 non-uniform FFT with jax-finufft.

    Arguments
    ---------
    ift :
        Array in real space. Assumes that the zero
        frequency component is in the center.
    Returns
    -------
    ft :
        Fourier transform of array.
    """
    ft = jnp.fft.fftn(jnp.fft.ifftshift(ift, axes=axes), axes=axes, **kwargs)

    return ft


@overload
def irfftn(
    ft: Complex[Array, "y_dim x_dim//2+1"],
    s: Optional[tuple[int, ...]] = None,
    axes: Optional[tuple[int, ...]] = None,
    **kwargs: Any,
) -> Float[Array, "y_dim x_dim"] | Float[Array, " *s"]: ...


@overload
def irfftn(
    ft: Complex[Array, "z_dim y_dim x_dim//2+1"],
    s: Optional[tuple[int, ...]] = None,
    axes: Optional[tuple[int, ...]] = None,
    **kwargs: Any,
) -> Float[Array, "z_dim y_dim x_dim"] | Float[Array, " *s"]: ...


def irfftn(
    ft: Complex[Array, "y_dim x_dim//2+1"] | Complex[Array, "z_dim y_dim x_dim//2+1"],
    s: Optional[tuple[int, ...]] = None,
    axes: Optional[tuple[int, ...]] = None,
    **kwargs: Any,
) -> (
    Float[Array, "y_dim x_dim"]
    | Float[Array, "z_dim y_dim x_dim"]
    | Float[Array, " *s"]
):
    """
    Helper routine to compute the inverse fourier transform
    from real input.

    Arguments
    ---------
    ft :
        Fourier transform array. Assumes that the zero
        frequency component is in the center.
    Returns
    -------
    ift :
        Inverse fourier transform.
    """
    ift = jnp.fft.fftshift(jnp.fft.irfftn(ft, s=s, axes=axes, **kwargs), axes=axes)

    return ift


@overload
def rfftn(
    ift: Float[Array, "y_dim x_dim"],
    axes: Optional[tuple[int, ...]] = None,
    **kwargs: Any,
) -> Complex[Array, "y_dim x_dim//2+1"]: ...


@overload
def rfftn(
    ift: Float[Array, "z_dim y_dim x_dim"],
    axes: Optional[tuple[int, ...]] = None,
    **kwargs: Any,
) -> Complex[Array, "z_dim y_dim x_dim//2+1"]: ...


def rfftn(
    ift: Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"],
    axes: Optional[tuple[int, ...]] = None,
    **kwargs: Any,
) -> Complex[Array, "y_dim x_dim//2+1"] | Complex[Array, "z_dim y_dim x_dim//2+1"]:
    """
    Helper routine to compute the fourier transform of an array
    to match the output of a type 1 non-uniform FFT.

    Arguments
    ---------
    ift :
        Array in real space. Assumes that the zero
        frequency component is in the center.
    Returns
    -------
    ft :
        Fourier transform of array.
    """
    ft = jnp.fft.rfftn(jnp.fft.ifftshift(ift, axes=axes), axes=axes, **kwargs)

    return ft



from ._lie_group_transforms import (
    AbstractLieGroupTransform as AbstractLieGroupTransform,
    apply_updates_with_lie_transform as apply_updates_with_lie_transform,
    SE3Transform as SE3Transform,
    SO3Transform as SO3Transform,
)
from ._transforms import (
    AbstractParameterTransform as AbstractParameterTransform,
    ComposedTransform as ComposedTransform,
    ExpTransform as ExpTransform,
    insert_transforms as insert_transforms,
    RescalingTransform as RescalingTransform,
    resolve_transforms as resolve_transforms,
)



import jax
import jax.numpy as jnp
import numpy as np
import pytest
from pycistem.core import AnglesAndShifts, CTF as cisCTF, Image

import cryojax.simulator as cs
from cryojax.coordinates import cartesian_to_polar, make_frequencies
from cryojax.image import irfftn, powerspectrum
from cryojax.io import read_array_with_spacing_from_mrc
from cryojax.simulator import CTF, EulerAnglePose


jax.config.update("jax_enable_x64", True)


@pytest.mark.parametrize(
    "defocus1,defocus2,asti_angle,kV,cs,ac,pixel_size",
    [
        (12000.0, 12000.0, 0.0, 300.0, 2.7, 0.07, 1.0),
        (12000.0, 12000.0, 0.0, 200.0, 0.01, 0.12, 1.3),
        (1200.0, 1200.0, 0.0, 300.0, 2.7, 0.07, 1.5),
        (24000.0, 12000.0, 30.0, 300.0, 2.7, 0.07, 0.9),
        (24000.0, 24000.0, 0.0, 300.0, 2.7, 0.07, 2.0),
        (9000.0, 7000.0, 180.0, 300.0, 2.7, 0.07, 1.0),
        (12000.0, 9000.0, 0.0, 200.0, 2.7, 0.07, 0.9),
        (12000.0, 12000.0, 60.0, 200.0, 2.7, 0.02, 0.75),
        (12000.0, 3895.0, 45.0, 200.0, 2.7, 0.07, 2.2),
    ],
)
def test_ctf_with_cistem(defocus1, defocus2, asti_angle, kV, cs, ac, pixel_size):
    """Test CTF model against cisTEM.

    Modified from https://github.com/jojoelfe/contrasttransferfunction"""
    shape = (512, 512)
    freqs = make_frequencies(shape, pixel_size)
    k_sqr, theta = cartesian_to_polar(freqs, square=True)
    # Compute cryojax CTF
    optics = CTF(
        defocus_u_in_angstroms=defocus1,
        defocus_v_in_angstroms=defocus2,
        astigmatism_angle=asti_angle,
        voltage_in_kilovolts=kV,
        spherical_aberration_in_mm=cs,
        amplitude_contrast_ratio=ac,
    )
    ctf = np.array(optics(freqs))
    # Compute cisTEM CTF
    cisTEM_optics = cisCTF(
        kV=kV,
        cs=cs,
        ac=ac,
        defocus1=defocus1,
        defocus2=defocus2,
        astig_angle=asti_angle,
        pixel_size=pixel_size,
    )
    cisTEM_ctf = np.vectorize(
        lambda k_sqr, theta: cisTEM_optics.Evaluate(k_sqr, theta)
    )(k_sqr.ravel() * pixel_size**2, theta.ravel()).reshape(freqs.shape[0:2])
    cisTEM_ctf[0, 0] = 0.0

    # Compute cryojax and cisTEM power spectrum
    radial_freqs = jnp.linalg.norm(freqs, axis=-1)
    spectrum1D, _ = powerspectrum(
        ctf, radial_freqs, pixel_size, k_max=1 / (2 * pixel_size)
    )
    cisTEM_spectrum1D, _ = powerspectrum(
        cisTEM_ctf, radial_freqs, pixel_size, k_max=1 / (2 * pixel_size)
    )

    np.testing.assert_allclose(ctf, cisTEM_ctf, atol=5e-2)
    np.testing.assert_allclose(spectrum1D, cisTEM_spectrum1D, atol=5e-3)


@pytest.mark.parametrize(
    "phi, theta, psi",
    [(10.0, 90.0, 170.0)],
    # [(10, 80, -20), (1.2, -90.5, 67), (-50, 62, -21)],
)
def test_euler_matrix_with_cistem(phi, theta, psi):
    """Test zyz rotation matrix"""
    # Hard code zyz rotation matrix from cisTEM convention
    phi_in_rad, theta_in_rad, psi_in_rad = [
        np.deg2rad(angle) for angle in [phi, theta, psi]
    ]
    matrix = np.zeros((3, 3))
    cos_phi = np.cos(phi_in_rad)
    sin_phi = np.sin(phi_in_rad)
    cos_theta = np.cos(theta_in_rad)
    sin_theta = np.sin(theta_in_rad)
    cos_psi = np.cos(psi_in_rad)
    sin_psi = np.sin(psi_in_rad)
    matrix[0, 0] = cos_phi * cos_theta * cos_psi - sin_phi * sin_psi
    matrix[1, 0] = sin_phi * cos_theta * cos_psi + cos_phi * sin_psi
    matrix[2, 0] = -sin_theta * cos_psi
    matrix[0, 1] = -cos_phi * cos_theta * sin_psi - sin_phi * cos_psi
    matrix[1, 1] = -sin_phi * cos_theta * sin_psi + cos_phi * cos_psi
    matrix[2, 1] = sin_theta * sin_psi
    matrix[0, 2] = sin_theta * cos_phi
    matrix[1, 2] = sin_theta * sin_phi
    matrix[2, 2] = cos_theta
    # Generate rotation that matches this rotation matrix
    pose = EulerAnglePose(view_phi=phi, view_theta=theta, view_psi=psi)
    np.testing.assert_allclose(pose.rotation.as_matrix(), matrix.T, atol=1e-12)


@pytest.mark.parametrize(
    "phi, theta, psi",
    [(10.0, 90.0, 170.0)],
)
def test_compute_projection_with_cistem(
    phi,
    theta,
    psi,
    sample_mrc_path,
    pixel_size,
):
    # cryojax
    real_voxel_grid, voxel_size = read_array_with_spacing_from_mrc(sample_mrc_path)
    potential = cs.FourierVoxelGridPotential.from_real_voxel_grid(
        real_voxel_grid, voxel_size
    )
    pose = cs.EulerAnglePose(view_phi=phi, view_theta=theta, view_psi=psi)
    integrator = cs.FourierSliceExtract()
    specimen = cs.Specimen(potential, integrator, pose)
    box_size = potential.shape[0]
    config = cs.ImageConfig((box_size, box_size), pixel_size)
    instrument = cs.Instrument(voltage_in_kilovolts=300.0)
    pipeline = cs.ImagePipeline(config, specimen, instrument)
    cryojax_projection = irfftn(
        pipeline.render(get_real=False).at[0, 0].set(0.0 + 0.0j)
        / np.sqrt(np.prod(config.shape)),
        s=config.padded_shape,
    )
    # pycistem
    pycistem_volume = _load_pycistem_template(sample_mrc_path, box_size)
    pycistem_angles = AnglesAndShifts()
    pycistem_angles.Init(phi, theta, psi, 0.0, 0.0)
    pycistem_model = _compute_projection(pycistem_volume, pycistem_angles, box_size)
    pycistem_projection = np.asarray(pycistem_model.real_values)

    np.testing.assert_allclose(cryojax_projection, pycistem_projection, atol=1e-5)


def _load_pycistem_template(filename, box_size):
    """Load pycistem template in fourier space."""
    volume = Image()
    volume.QuickAndDirtyReadSlices(filename, 1, box_size)
    volume.ForwardFFT(True)
    volume.ZeroCentralPixel()
    volume.SwapRealSpaceQuadrants()

    return volume


def _compute_projection(volume, angles, box_size):
    """Compute scattering projection of pycistem volume."""
    projection = Image()
    projection.Allocate(box_size, box_size, False)

    volume.ExtractSlice(projection, angles, 1.0, False)
    projection.PhaseShift(angles.ReturnShiftX(), angles.ReturnShiftY(), 0.0)

    projection.SwapRealSpaceQuadrants()
    projection.BackwardFFT()

    return projection



"""Unit conversions."""

import jax.numpy as jnp
from jaxtyping import Array, Float


def convert_keV_to_angstroms(
    energy_in_keV: Float[Array, ""] | float,
) -> Float[Array, ""]:
    """Get the relativistic electron wavelength at a given accelerating voltage."""
    energy_in_eV = 1000.0 * energy_in_keV  # keV to eV
    return jnp.asarray(12.2643 / (energy_in_eV + 0.97845e-6 * energy_in_eV**2) ** 0.5)



"""
Image normalization routines.
"""

from typing import Optional

import jax.numpy as jnp
from jaxtyping import Array, Complex, Float, Inexact


def normalize_image(
    image: Inexact[Array, "y_dim x_dim"],
    *,
    is_real: bool = True,
    half_space: bool = True,
    shape_in_real_space: Optional[tuple[int, int]] = None,
) -> Inexact[Array, "y_dim x_dim"]:
    """
    Normalize so that the image is mean 0
    and standard deviation 1 in real space.
    """
    return rescale_image(
        image,
        1.0,
        0.0,
        is_real=is_real,
        half_space=half_space,
        shape_in_real_space=shape_in_real_space,
    )


def rescale_image(
    image: Inexact[Array, "y_dim x_dim"],
    std: float | Float[Array, ""],
    mean: float | Float[Array, ""],
    *,
    is_real: bool = True,
    half_space: bool = True,
    shape_in_real_space: Optional[tuple[int, int]] = None,
) -> Inexact[Array, "y_dim x_dim"]:
    """Normalize so that the image is mean mu
    and standard deviation N in real space.

    Parameters
    ----------
    image : `Array`, shape `(N1, N2)`
        The image in either real or Fourier space.
        If in Fourier space, the zero frequency
        component should be in the corner of the image.
        It should also be on the half space.
    std : `float`
        Intensity standard deviation.
    mean : `float`
        Intensity offset.
    is_real : `bool`
        If ``True``, the given ``image`` is in real
        space. If ``False``, it is in Fourier space.

    Returns
    -------
    rescaled_image : `Array`, shape `(N1, N2)`
        Image rescaled to have mean ``mu`` and standard
        deviation ``N``.
    """
    image = jnp.asarray(image)
    N1, N2 = image.shape
    # First normalize image to zero mean and unit standard deviation
    if is_real:
        normalized_image = (image - image.mean()) / image.std()
        rescaled_image = std * normalized_image + mean
    else:
        if half_space:
            N_modes = (
                N1 * (2 * N2 - 1)
                if shape_in_real_space is None
                else shape_in_real_space[0] * shape_in_real_space[1]
            )
        else:
            N_modes = N1 * N2
        _, measured_std = compute_mean_and_std_from_fourier_image(
            image, half_space=half_space, shape_in_real_space=shape_in_real_space
        )
        normalized_image = image.at[0, 0].set(0.0) / measured_std
        rescaled_image = (normalized_image * std).at[0, 0].set(mean * N_modes)
    return rescaled_image


def compute_mean_and_std_from_fourier_image(
    fourier_image: Complex[Array, "y_dim x_dim"],
    *,
    half_space: bool = True,
    shape_in_real_space: Optional[tuple[int, int]] = None,
) -> tuple[Float[Array, ""], Float[Array, ""]]:
    """Compute the mean and standard deviation in real space from
    an image in fourier space.
    """
    N1, N2 = fourier_image.shape
    if half_space:
        N_modes = (
            N1 * (2 * N2 - 1)
            if shape_in_real_space is None
            else shape_in_real_space[0] * shape_in_real_space[1]
        )
    else:
        N_modes = N1 * N2
    # The mean is just the zero mode divided by the number of modes
    mean = fourier_image[0, 0] / N_modes
    # The standard deviation is square root norm squared of the zero mean image
    std = (
        jnp.sqrt(
            jnp.sum(jnp.abs(fourier_image[:, 0]) ** 2)
            - jnp.abs(fourier_image[0, 0]) ** 2
            + 2 * jnp.sum(jnp.abs(fourier_image[:, 1:]) ** 2)
        )
        if half_space
        else jnp.linalg.norm(fourier_image.at[0, 0].set(0.0))
    ) / N_modes

    return (mean, std)



"""
Routines to compute radial averages of images.
"""

from functools import partial
from typing import overload

import jax
import jax.numpy as jnp
from jaxtyping import Array, Float, Inexact


@overload
def radial_average(
    image: Inexact[Array, "y_dim x_dim"],
    radial_grid: Float[Array, "y_dim x_dim"],
    bins: Float[Array, " n_bins"],
) -> Inexact[Array, " n_bins"]: ...


@overload
def radial_average(
    image: Inexact[Array, "z_dim y_dim x_dim"],
    radial_grid: Float[Array, "z_dim y_dim x_dim"],
    bins: Float[Array, " n_bins"],
) -> Inexact[Array, " n_bins"]: ...


@overload
def radial_average(
    image: Inexact[Array, "y_dim x_dim"],
    radial_grid: Float[Array, "y_dim x_dim"],
    bins: Float[Array, " n_bins"],
    *,
    to_grid: bool = False,
    interpolation_mode: str = "nearest",
) -> tuple[Inexact[Array, " n_bins"], Inexact[Array, "y_dim x_dim"]]: ...


@overload
def radial_average(
    image: Inexact[Array, "z_dim y_dim x_dim"],
    radial_grid: Float[Array, "z_dim y_dim x_dim"],
    bins: Float[Array, " n_bins"],
    *,
    to_grid: bool = False,
    interpolation_mode: str = "nearest",
) -> tuple[Inexact[Array, " n_bins"], Inexact[Array, "z_dim y_dim x_dim"]]: ...


@partial(jax.jit, static_argnames=["to_grid", "interpolation_mode"])
def radial_average(
    image: Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"],
    radial_grid: Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"],
    bins: Float[Array, " n_bins"],
    *,
    to_grid: bool = False,
    interpolation_mode: str = "nearest",
) -> (
    Inexact[Array, " n_bins"]
    | tuple[
        Inexact[Array, " n_bins"],
        Inexact[Array, "y_dim x_dim"] | Inexact[Array, "z_dim y_dim x_dim"],
    ]
):
    """
    Radially average vectors r with a given magnitude
    coordinate system |r|.

    Arguments
    ---------
    image :
        Two-dimensional image.
    radial_grid :
        Radial coordinate system of image.
    bins :
        Radial bins for averaging. These
        must be evenly spaced.
    to_grid :
        If ``False``, evalulate the spectrum as a 1D
        profile. Otherwise, evaluate the spectrum on the
        grid.
    interpolation_mode :
        If ``"linear"``, evaluate on the grid using linear
        interpolation. If ``False``,

    Returns
    -------
    average :
        Radial average of image.
    """
    bins = jnp.asarray(bins)
    # Discretize the radial grid
    digitized_radial_grid = jnp.digitize(radial_grid, bins, right=True)
    # Compute the radial profile as the average value of the image in each bin
    average_as_profile = jnp.bincount(
        digitized_radial_grid.ravel(),
        weights=image.ravel(),
        length=bins.size,
    ) / jnp.bincount(digitized_radial_grid.ravel(), length=bins.size)
    # Interpolate to a grid or return the profile
    if to_grid:
        if interpolation_mode == "nearest":
            average_as_grid = jnp.take(
                average_as_profile,
                digitized_radial_grid,
                mode="clip",
            )
        elif interpolation_mode == "linear":
            average_as_grid = jnp.interp(
                radial_grid.ravel(),
                bins,
                average_as_profile,
            ).reshape(radial_grid.shape)
        else:
            raise ValueError(
                f"interpolation_mode = {interpolation_mode} not supported."
            )
        return average_as_profile, average_as_grid
    else:
        return average_as_profile



"""Pytrees that represent particle stacks."""

from equinox import AbstractVar, Module
from jaxtyping import Array, Inexact


class AbstractParticleStack(Module, strict=True):
    """An abstraction of a particle stack, represented by a pytree
    that stores an image stack.

    Subclasses should also include parameters included in the
    image formation model, typically represented with `cryojax` objects.
    """

    image_stack: AbstractVar[Inexact[Array, "... y_dim x_dim"]]



"""
Methods for integrating the scattering potential onto the exit plane.
"""

from abc import abstractmethod
from typing import Generic, TypeVar

from equinox import Module
from jaxtyping import Array, Complex, Float

from .._config import ImageConfig
from .._potential import AbstractScatteringPotential


ScatteringPotentialT = TypeVar(
    "ScatteringPotentialT", bound="AbstractScatteringPotential"
)


class AbstractPotentialIntegrator(Module, Generic[ScatteringPotentialT], strict=True):
    """Base class for a method of integrating the scattering
    potential to a set of phase shifts the exit plane."""

    @abstractmethod
    def __call__(
        self,
        potential: ScatteringPotentialT,
        wavelength_in_angstroms: Float[Array, ""],
        config: ImageConfig,
    ) -> Complex[Array, "{config.padded_y_dim} {config.padded_x_dim//2+1}"]:
        """Compute the scattering potential in the exit plane at
        the `ImageConfig` settings.

        **Arguments:**

        - `potential`: The scattering potential representation.
        - `wavelength_in_angstroms`: The wavelength of the electron beam.
        - `config`: The configuration of the resulting image.
        """
        raise NotImplementedError



from functools import partial

import equinox as eqx
import jax
import jax.numpy as jnp
import jax.tree_util as jtu
from jaxtyping import Array, Float

import cryojax.simulator as cs
from cryojax.constants import convert_keV_to_angstroms
from cryojax.coordinates import (
    AbstractCoordinates,
    CoordinateGrid,
    CoordinateList,
    FrequencySlice,
)


def test_voxel_electron_potential_loaders():
    real_voxel_grid = jnp.zeros((10, 10, 10), dtype=float)
    voxel_size = 1.1
    fourier_potential = cs.FourierVoxelGridPotential.from_real_voxel_grid(
        real_voxel_grid, voxel_size=voxel_size
    )
    real_potential = cs.RealVoxelGridPotential.from_real_voxel_grid(
        real_voxel_grid, voxel_size=voxel_size
    )
    cloud_potential = cs.RealVoxelCloudPotential.from_real_voxel_grid(
        real_voxel_grid, voxel_size=voxel_size
    )
    for potential in [real_potential, fourier_potential, cloud_potential]:
        assert potential.voxel_size == jnp.asarray(voxel_size)

    assert isinstance(
        fourier_potential.wrapped_frequency_slice_in_pixels, FrequencySlice
    )
    assert isinstance(
        fourier_potential.wrapped_frequency_slice_in_pixels.get(),
        Float[Array, "1 _ _ 3"],
    )
    assert isinstance(real_potential.wrapped_coordinate_grid_in_pixels, CoordinateGrid)
    assert isinstance(
        real_potential.wrapped_coordinate_grid_in_pixels.get(), Float[Array, "_ _ _ 3"]
    )
    assert isinstance(cloud_potential.wrapped_coordinate_list_in_pixels, CoordinateList)
    assert isinstance(
        cloud_potential.wrapped_coordinate_list_in_pixels.get(), Float[Array, "_ 3"]
    )


def test_electron_potential_vmap(potential, integrator, config):
    filter_spec = jtu.tree_map(
        lambda x: not isinstance(x, AbstractCoordinates),
        potential,
        is_leaf=lambda x: isinstance(x, AbstractCoordinates),
    )
    # Add a leading dimension to scattering potential leaves
    potential = jtu.tree_map(
        lambda spec, x: jnp.expand_dims(x, axis=0) if spec else x,
        filter_spec,
        potential,
        is_leaf=lambda x: isinstance(x, AbstractCoordinates),
    )
    vmap, novmap = eqx.partition(potential, filter_spec)

    @partial(jax.vmap, in_axes=[0, None, None, None])
    def compute_image_stack(vmap, novmap, integrator, config):
        wavelength_in_angstroms = convert_keV_to_angstroms(300.0)
        potential = eqx.combine(vmap, novmap)
        return integrator(potential, wavelength_in_angstroms, config)

    # vmap over first axis
    image_stack = compute_image_stack(vmap, novmap, integrator, config)
    assert image_stack.shape[:1] == (1,)


def test_electron_potential_vmap_with_pipeline(potential, pose, integrator, config):
    instrument = cs.Instrument(voltage_in_kilovolts=300.0)
    pipeline = cs.ImagePipeline(
        config, cs.Specimen(potential, integrator, pose), instrument
    )

    def is_potential_leaves_without_coordinates(element):
        if isinstance(element, cs.AbstractScatteringPotential):
            return jtu.tree_map(
                lambda x: not isinstance(x, AbstractCoordinates),
                potential,
                is_leaf=lambda x: isinstance(x, AbstractCoordinates),
            )
        else:
            return False

    # Get filter spec for scattering potential
    filter_spec = jtu.tree_map(
        is_potential_leaves_without_coordinates,
        pipeline,
        is_leaf=lambda x: isinstance(x, cs.AbstractScatteringPotential),
    )
    # Add a leading dimension to scattering potential leaves
    pipeline = jtu.tree_map(
        lambda spec, x: jnp.expand_dims(x, axis=0) if spec else x,
        filter_spec,
        pipeline,
    )
    vmap, novmap = eqx.partition(pipeline, filter_spec)

    @partial(jax.vmap, in_axes=[0, None])
    def compute_image_stack(vmap, novmap):
        pipeline = eqx.combine(vmap, novmap)
        return pipeline.render()

    # vmap over first axis
    image_stack = compute_image_stack(vmap, novmap)
    assert image_stack.shape[:1] == (1,)



"""Cryojax compatibility with [RELION](https://relion.readthedocs.io/en/release-5.0/)."""

import dataclasses
import pathlib
from typing import Any, Callable, final

import equinox as eqx
import jax.numpy as jnp
import mrcfile
import numpy as np
import pandas as pd
from jaxtyping import Array, Float, Int

from ..io import read_and_validate_starfile
from ..simulator import CTF, EulerAnglePose, ImageConfig
from ._dataset import AbstractDataset
from ._particle_stack import AbstractParticleStack


RELION_REQUIRED_OPTICS_KEYS = [
    "rlnImageSize",
    "rlnVoltage",
    "rlnImagePixelSize",
    "rlnSphericalAberration",
    "rlnAmplitudeContrast",
]
RELION_REQUIRED_PARTICLE_KEYS = [
    "rlnDefocusU",
    "rlnDefocusV",
    "rlnDefocusAngle",
    "rlnPhaseShift",
    "rlnImageName",
]


class RelionParticleStack(AbstractParticleStack):
    """A particle stack with information imported from
    [RELION](https://relion.readthedocs.io/en/release-5.0/).
    """

    image_stack: Float[Array, "... y_dim x_dim"]
    config: ImageConfig
    pose: EulerAnglePose
    ctf: CTF

    def __init__(
        self,
        image_stack: Float[Array, "... y_dim x_dim"],
        config: ImageConfig,
        pose: EulerAnglePose,
        ctf: CTF,
    ):
        # Set image stack and config as is
        self.image_stack = jnp.asarray(image_stack)
        self.config = config
        # Set CTF using the defocus offset in the EulerAnglePose
        self.ctf = eqx.tree_at(
            lambda ctf: (ctf.defocus_u_in_angstroms, ctf.defocus_v_in_angstroms),
            ctf,
            (
                ctf.defocus_u_in_angstroms + pose.offset_z_in_angstroms,
                ctf.defocus_v_in_angstroms + pose.offset_z_in_angstroms,
            ),
        )
        # Set defocus offset to zero
        self.pose = eqx.tree_at(
            lambda pose: pose.offset_z_in_angstroms, pose, jnp.asarray(0.0)
        )


RelionParticleStack.__init__.__doc__ = """**Arguments:**

- `image_stack`: The stack of images. The shape of this array
                 is a leading batch dimension followed by the shape
                 of an image in the stack.
- `config`: The image configuration. Any subset of pytree leaves may
            have a batch dimension.
- `pose`: The pose, represented by euler angles. Any subset of pytree leaves may
          have a batch dimension. Upon instantiation, `pose.offset_z_in_angstroms`
          is set to zero.
- `ctf`: The contrast transfer function. Any subset of pytree leaves may
         have a batch dimension. Upon instantiation, `ctf.defocus_u_in_angstroms`
         is set to `ctf.defocus_u_in_angstroms + pose.offset_z_in_angstroms` (and
         also for `ctf.defocus_v_in_angstroms`).
"""


def default_relion_make_config(
    shape: tuple[int, int], pixel_size: float | Float[np.ndarray, "..."], **kwargs: Any
):
    return ImageConfig(shape, jnp.asarray(pixel_size), **kwargs)


@dataclasses.dataclass(frozen=True)
class RelionDataset(AbstractDataset):
    """A dataset that wraps a Relion particle stack in
    [STAR](https://relion.readthedocs.io/en/latest/Reference/Conventions.html) format.
    """

    path_to_relion_project: pathlib.Path
    data_blocks: dict[str, pd.DataFrame]

    make_config: Callable[
        [tuple[int, int], float | Float[np.ndarray, "..."]], ImageConfig
    ]

    @final
    def __init__(
        self,
        path_to_starfile: str | pathlib.Path,
        path_to_relion_project: str | pathlib.Path,
        make_config: Callable[
            [tuple[int, int], float | Float[np.ndarray, "..."]], ImageConfig
        ] = default_relion_make_config,
    ):
        """**Arguments:**

        - `path_to_starfile`: The path to the Relion STAR file.
        - `path_to_relion_project`: The path to the Relion project directory.
        """
        data_blocks = read_and_validate_starfile(path_to_starfile)
        _validate_relion_data_blocks(data_blocks)
        object.__setattr__(self, "data_blocks", data_blocks)
        object.__setattr__(
            self, "path_to_relion_project", pathlib.Path(path_to_relion_project)
        )
        object.__setattr__(self, "make_config", make_config)

    @final
    def __getitem__(
        self, index: int | slice | Int[np.ndarray, ""] | Int[np.ndarray, " N"]
    ) -> RelionParticleStack:
        # Load particle data and optics group
        n_rows = self.data_blocks["particles"].shape[0]
        index_error_msg = lambda idx: (
            "The index at which the `RelionDataset` was accessed was out of bounds! "
            f"The number of rows in the dataset is {n_rows}, but you tried to "
            f"access the index {idx}."
        )
        # pandas has bad error messages for its indexing
        if isinstance(index, (int, Int[np.ndarray, ""])):
            if index > n_rows - 1:
                raise IndexError(index_error_msg(index))
        elif isinstance(index, slice):
            if index.start > n_rows - 1:
                raise IndexError(index_error_msg(index.start))
        elif isinstance(index, np.ndarray):
            pass  # catch exceptions later
        else:
            raise IndexError(
                f"Indexing with the type {type(index)} is not supported by "
                "`RelionDataset`. Indexing by integers is supported, one-dimensional "
                "fancy indexing is supported, and numpy-array indexing is supported. "
                "For example, like `particle = dataset[0]`, "
                "`particle_stack = dataset[0:5]`, "
                "or `particle_stack = dataset[np.array([1, 4, 3, 2])]`."
            )
        try:
            particle_blocks = self.data_blocks["particles"].iloc[index]
        except Exception:
            raise IndexError(
                "Error when indexing the `pandas.Dataframe` for the particle stack "
                "from the `starfile.read` output."
            )
        optics_group = self.data_blocks["optics"].iloc[0]
        # Load particle image stack rlnImageName
        image_stack_index_and_name_series_or_str = particle_blocks["rlnImageName"]
        if isinstance(image_stack_index_and_name_series_or_str, str):
            # In this block, the user most likely used standard indexing, like
            # `dataset = RelionDataset(...); particle_stack = dataset[1]`
            image_stack_index_and_name_str = image_stack_index_and_name_series_or_str
            # ... split the whole string into its image index and filename
            relion_particle_index, image_stack_filename = (
                image_stack_index_and_name_str.split("@")
            )
            # ... create full path to the image stack
            path_to_image_stack = pathlib.Path(
                self.path_to_relion_project, image_stack_filename
            )
            # ... relion convention starts indexing at 1, not 0
            particle_index = np.asarray(relion_particle_index, dtype=int) - 1
        elif isinstance(image_stack_index_and_name_series_or_str, pd.Series):
            # In this block, the user most likely used fancy indexing, like
            # `dataset = RelionDataset(...); particle_stack = dataset[1:10]`
            image_stack_index_and_name_series = image_stack_index_and_name_series_or_str
            # ... split the pandas.Series into a pandas.DataFrame with two columns:
            # one for the image index and another for the filename
            image_stack_index_and_name_dataframe = (
                image_stack_index_and_name_series.str.split("@", expand=True)
            )
            # ... get a pandas.Series for each the index and the filename
            relion_particle_index, image_stack_filename = [
                image_stack_index_and_name_dataframe[column]
                for column in image_stack_index_and_name_dataframe.columns
            ]
            # ... multiple filenames in the same STAR file is not supported with
            # fancy indexing
            if image_stack_filename.nunique() != 1:
                raise ValueError(
                    "Found multiple image stack filenames when reading "
                    "STAR file rows. This is most likely because you tried to "
                    "use fancy indexing with multiple image stack filenames "
                    "in the same STAR file. If a STAR file refers to multiple image "
                    "stack filenames, fancy indexing is not supported. For example, "
                    "this will raise an error: `dataset = RelionDataset(...); "
                    "particle_stack = dataset[1:10]`."
                )
            # ... create full path to the image stack
            path_to_image_stack = pathlib.Path(
                self.path_to_relion_project,
                np.asarray(image_stack_filename, dtype=object)[0],
            )
            # ... relion convention starts indexing at 1, not 0
            particle_index = (
                np.asarray(relion_particle_index.astype(int), dtype=int) - 1
            )
        else:
            raise IOError(
                "Could not read `rlnImageName` in STAR file for `RelionDataset` "
                f"index equal to {index}."
            )
        with mrcfile.mmap(path_to_image_stack, mode="r", permissive=True) as mrc:
            image_stack = np.asarray(mrc.data[particle_index])  # type: ignore
        # Read metadata into a RelionParticleStack
        # ... particle data
        defocus_u_in_angstroms = np.asarray(particle_blocks["rlnDefocusU"])
        defocus_v_in_angstroms = np.asarray(particle_blocks["rlnDefocusV"])
        astigmatism_angle = np.asarray(particle_blocks["rlnDefocusAngle"])
        phase_shift = np.asarray(particle_blocks["rlnPhaseShift"])
        # ... optics group data
        image_size = np.asarray(optics_group["rlnImageSize"])
        pixel_size = np.asarray(optics_group["rlnImagePixelSize"])
        voltage_in_kilovolts = np.asarray(optics_group["rlnVoltage"])
        spherical_aberration_in_mm = np.asarray(optics_group["rlnSphericalAberration"])
        amplitude_contrast_ratio = np.asarray(optics_group["rlnAmplitudeContrast"])
        # ... create cryojax objects
        config = self.make_config((int(image_size), int(image_size)), pixel_size)
        ctf = CTF(
            defocus_u_in_angstroms,
            defocus_v_in_angstroms,
            astigmatism_angle,
            voltage_in_kilovolts,
            spherical_aberration_in_mm,
            amplitude_contrast_ratio,
            phase_shift,
        )
        pose = EulerAnglePose()
        # ... values for the pose are optional, so look to see if
        # each key is present
        particle_keys = particle_blocks.keys()
        pose_parameter_names_and_values = []
        if "rlnOriginXAngst" in particle_keys:
            pose_parameter_names_and_values.append(
                ("offset_x_in_angstroms", particle_blocks["rlnOriginXAngst"])
            )
        if "rlnOriginYAngst" in particle_keys:
            pose_parameter_names_and_values.append(
                ("offset_y_in_angstroms", particle_blocks["rlnOriginYAngst"])
            )
        if "rlnAngleRot" in particle_keys:
            pose_parameter_names_and_values.append(
                ("view_phi", particle_blocks["rlnAngleRot"])
            )
        if "rlnAngleTilt" in particle_keys:
            pose_parameter_names_and_values.append(
                ("view_theta", particle_blocks["rlnAngleTilt"])
            )
        elif "rlnAngleTiltPrior" in particle_keys:  # support for helices
            pose_parameter_names_and_values.append(
                ("view_theta", particle_blocks["rlnAngleTiltPrior"])
            )
        if "rlnAnglePsi" in particle_keys:
            # Relion uses -999.0 as a placeholder for an un-estimated in-plane
            # rotation
            if isinstance(particle_blocks["rlnAnglePsi"], pd.Series):
                # ... check if all values are equal to -999.0. If so, just
                # replace the whole pandas.Series with 0.0
                if (
                    particle_blocks["rlnAnglePsi"].nunique() == 1
                    and particle_blocks["rlnAnglePsi"][0] == -999.0
                ):
                    particle_blocks_for_psi = 0.0
                # ... otherwise, replace -999.0 values with 0.0
                else:
                    particle_blocks_for_psi = particle_blocks["rlnAnglePsi"].where(
                        lambda x: x != -999.0, 0.0
                    )
            else:
                # ... if the column is just equal to a float, then
                # directly check if it is equal to -999.0
                particle_blocks_for_psi = (
                    0.0
                    if particle_blocks["rlnAnglePsi"] == -999.0
                    else particle_blocks["rlnAnglePsi"]
                )
            pose_parameter_names_and_values.append(
                ("view_psi", particle_blocks_for_psi)
            )
        elif "rlnAnglePsiPrior" in particle_keys:  # support for helices
            pose_parameter_names_and_values.append(
                ("view_psi", particle_blocks["rlnAnglePsiPrior"])
            )
        pose_parameter_names, pose_parameter_values = tuple(
            zip(*pose_parameter_names_and_values)
        )
        # ... fill the EulerAnglePose will keys that are present. if they are not
        # present, keep the default values in the `pose = EulerAnglePose()`
        # instantiation
        pose = eqx.tree_at(
            lambda p: tuple([getattr(p, name) for name in pose_parameter_names]),
            pose,
            tuple([jnp.asarray(value) for value in pose_parameter_values]),
        )

        return RelionParticleStack(jnp.asarray(image_stack), config, pose, ctf)

    @final
    def __len__(self) -> int:
        return len(self.data_blocks["particles"])


@dataclasses.dataclass(frozen=True)
class HelicalRelionDataset(AbstractDataset):
    """A wrapped `RelionDataset` to read helical tubes.

    In particular, a `HelicalRelionDataset` indexes one
    helical filament at a time. For example, after manual
    particle picking in RELION, we can index a particular filament
    with

    ```python
    # Read in a STAR file particle stack
    dataset = RelionDataset(...)
    helical_dataset = HelicalRelionDataset(dataset)
    # ... get a particle stack for a filament
    particle_stack_for_a_filament = helical_dataset[0]
    # ... get a particle stack for another filament
    particle_stack_for_another_filament = helical_dataset[1]
    ```

    Unlike a `RelionDataset`, a `HelicalRelionDataset` does not
    support fancy indexing.
    """

    dataset: RelionDataset

    @final
    def __init__(
        self,
        dataset: RelionDataset,
    ):
        """**Arguments:**

        - `dataset`: The wrappped `RelionDataset`. This will be slightly
                     modified to read one helix at a time.
        """
        _validate_helical_relion_data_blocks(dataset.data_blocks)
        object.__setattr__(self, "dataset", dataset)

    @final
    def __getitem__(
        self, filament_index: int | Int[np.ndarray, ""]
    ) -> RelionParticleStack:
        if not isinstance(filament_index, (int, Int[np.ndarray, ""])):
            raise IndexError(
                "When indexing a `HelicalRelionDataset`, only "
                f"python or numpy-like integer indices are supported, such as "
                "`helical_particle_stack = helical_dataset[3]`. "
                f"Got index {filament_index} of type {type(filament_index)}."
            )
        # Read all images at a particular rlnHelicalTubeID
        particle_dataframe = self.dataset.data_blocks["particles"]
        # ... make sure the index is not out of bounds
        n_filaments = particle_dataframe["rlnHelicalTubeID"].max()
        if filament_index + 1 > n_filaments:
            raise ValueError(
                "The index at which the `HelicalRelionDataset` was accessed was out of "
                f"bounds! The number of filaments in the dataset is {n_filaments}, but "
                f"you tried to access the index {filament_index}."
            )
        # .. get the indices for a filament
        particle_indices = np.squeeze(
            np.argwhere(particle_dataframe["rlnHelicalTubeID"] == filament_index + 1)
        )
        # ... access the particle stack at these indices
        dataset = self.dataset[particle_indices]
        return dataset

    @final
    def __len__(self) -> int:
        return len(self.dataset)


def _validate_relion_data_blocks(data_blocks: dict[str, pd.DataFrame]):
    if "particles" not in data_blocks.keys():
        raise ValueError("Missing key 'particles' in `starfile.read` output.")
    else:
        if not set(RELION_REQUIRED_PARTICLE_KEYS).issubset(
            set(data_blocks["particles"].keys())
        ):
            raise ValueError(
                "Missing required keys in starfile 'particles' group. "
                f"Required keys are {RELION_REQUIRED_PARTICLE_KEYS}."
            )
    if "optics" not in data_blocks.keys():
        raise ValueError("Missing key 'optics' in `starfile.read` output.")
    else:
        if not set(RELION_REQUIRED_OPTICS_KEYS).issubset(
            set(data_blocks["optics"].keys())
        ):
            raise ValueError(
                "Missing required keys in starfile 'optics' group. "
                f"Required keys are {RELION_REQUIRED_OPTICS_KEYS}."
            )


def _validate_helical_relion_data_blocks(data_blocks: dict[str, pd.DataFrame]):
    particle_data_blocks = data_blocks["particles"]
    if "rlnHelicalTubeID" not in particle_data_blocks.columns:
        raise ValueError(
            "Missing column 'rlnHelicalTubeID' in `starfile.read` output. "
            "This column must be present when using a `HelicalRelionDataset`."
        )



"""
Abstraction of the electron microscope. This includes models
for the optics, electron dose, and detector.
"""

from typing import Optional

import jax.numpy as jnp
from equinox import field, Module
from jaxtyping import Array, Complex, Float, PRNGKeyArray

from ..constants import convert_keV_to_angstroms
from ..core import error_if_not_positive
from ._config import ImageConfig
from ._detector import AbstractDetector
from ._dose import ElectronDose
from ._optics import AbstractOptics


class Instrument(Module, strict=True):
    """An abstraction of an electron microscope.

    **Attributes:**

    - `voltage_in_kilovolts`: The accelerating voltage of the
                              instrument in kilovolts (kV).
    - `optics`: The model for the instrument optics.
    - `dose`: The model for the exposure to electrons
              during image formation.
    - `detector` : The model of the detector.
    """

    voltage_in_kilovolts: Float[Array, ""] = field(converter=error_if_not_positive)
    dose: Optional[ElectronDose]
    optics: Optional[AbstractOptics]
    detector: Optional[AbstractDetector]

    def __init__(
        self,
        voltage_in_kilovolts: float | Float[Array, ""],
        *,
        dose: Optional[ElectronDose] = None,
        optics: Optional[AbstractOptics] = None,
        detector: Optional[AbstractDetector] = None,
    ):
        if (optics is None or dose is None) and isinstance(detector, AbstractDetector):
            raise AttributeError(
                "Cannot set Instrument.detector without passing an AbstractOptics and "
                "an ElectronDose."
            )
        self.voltage_in_kilovolts = jnp.asarray(voltage_in_kilovolts)
        self.optics = optics
        self.dose = dose
        self.detector = detector

    @property
    def wavelength_in_angstroms(self) -> Float[Array, ""]:
        return convert_keV_to_angstroms(self.voltage_in_kilovolts)

    def propagate_to_detector_plane(
        self,
        fourier_phase_at_exit_plane: Complex[
            Array, "{config.padded_y_dim} {config.padded_x_dim//2+1}"
        ],
        config: ImageConfig,
        defocus_offset: Float[Array, ""] | float = 0.0,
    ) -> (
        Complex[Array, "{config.padded_y_dim} {config.padded_x_dim//2+1}"]
        | Complex[Array, "{config.padded_y_dim} {config.padded_x_dim}"]
    ):
        if self.optics is None:
            raise AttributeError(
                "Tried to call `Instrument.propagate_to_detector_plane`, "
                "but the `Instrument`'s optics model is `None`. This "
                "is not allowed!"
            )
        """Propagate the scattering potential with the optics model."""
        fourier_contrast_at_detector_plane = self.optics(
            fourier_phase_at_exit_plane,
            config,
            self.wavelength_in_angstroms,
            defocus_offset=defocus_offset,
        )

        return fourier_contrast_at_detector_plane

    def compute_fourier_squared_wavefunction(
        self,
        fourier_contrast_at_detector_plane: (
            Complex[Array, "{config.padded_y_dim} {config.padded_x_dim//2+1}"]
            | Complex[Array, "{config.padded_y_dim} {config.padded_x_dim}"]
        ),
        config: ImageConfig,
    ) -> Complex[Array, "{config.padded_y_dim} {config.padded_x_dim//2+1}"]:
        """Compute the squared wavefunction at the detector plane, given the
        contrast.
        """
        N1, N2 = config.padded_shape
        if self.optics is None:
            raise AttributeError(
                "Tried to call `compute_fourier_squared_wavefunction`, "
                "but the `Instrument`'s optics model is `None`. This "
                "is not allowed!"
            )
        elif self.optics.is_linear:
            # ... compute the squared wavefunction directly from the image contrast
            # as |psi|^2 = 1 + 2C.
            fourier_contrast_at_detector_plane = fourier_contrast_at_detector_plane
            fourier_squared_wavefunction_at_detector_plane = (
                (2 * fourier_contrast_at_detector_plane).at[0, 0].add(1.0 * N1 * N2)
            )
            return fourier_squared_wavefunction_at_detector_plane
        else:
            raise NotImplementedError(
                "Functionality for AbstractOptics.is_linear = False not supported."
            )

    def compute_expected_electron_events(
        self,
        fourier_squared_wavefunction_at_detector_plane: Complex[
            Array, "{config.padded_y_dim} {config.padded_x_dim//2+1}"
        ],
        config: ImageConfig,
    ) -> Complex[Array, "{config.padded_y_dim} {config.padded_x_dim//2+1}"]:
        """Compute the expected electron events from the detector."""
        if self.detector is None:
            raise AttributeError(
                "Tried to call `Instrument.compute_expected_electron_events`, "
                "but the `Instrument`'s detector model is `None`. This "
                "is not allowed!"
            )
        fourier_expected_electron_events = self.detector(
            fourier_squared_wavefunction_at_detector_plane,
            config,
            self.dose.electrons_per_angstrom_squared,
            key=None,
        )

        return fourier_expected_electron_events

    def measure_detector_readout(
        self,
        key: PRNGKeyArray,
        fourier_squared_wavefunction_at_detector_plane: Complex[
            Array, "{config.padded_y_dim} {config.padded_x_dim//2+1}"
        ],
        config: ImageConfig,
    ) -> Complex[Array, "{config.padded_y_dim} {config.padded_x_dim//2+1}"]:
        """Measure the readout from the detector."""
        if self.detector is None:
            raise AttributeError(
                "Tried to call `Instrument.measure_detector_readout`, "
                "but the `Instrument`'s detector model is `None`. This "
                "is not allowed!"
            )
        fourier_detector_readout = self.detector(
            fourier_squared_wavefunction_at_detector_plane,
            config,
            self.dose.electrons_per_angstrom_squared,
            key,
        )

        return fourier_detector_readout



"""
Representations of conformational variables.
"""

from typing import Any

from equinox import AbstractVar, field, Module
from jaxtyping import Array, Int

from ..core import error_if_negative


class AbstractConformation(Module, strict=True):
    """
    A conformational variable wrapped in a Module.
    """

    value: AbstractVar[Any]


class DiscreteConformation(AbstractConformation, strict=True):
    """
    A conformational variable wrapped in a Module.
    """

    value: Int[Array, ""] = field(converter=error_if_negative)



from ._load_atoms import (
    default_form_factor_params as default_form_factor_params,
    get_form_factor_params as get_form_factor_params,
)
from ._unit_conversions import convert_keV_to_angstroms as convert_keV_to_angstroms



import jax
import jax.numpy as jnp
import numpy as np
import pytest

from cryojax.image import fftn, ifftn


jax.config.update("jax_enable_x64", True)


@pytest.mark.parametrize("model", ["noisy_model", "noiseless_model"])
def test_fft(model, request):
    model = request.getfixturevalue(model)
    image = model.render()
    random = jnp.asarray(np.random.randn(*image.shape))
    # Run tests with jnp.fft and random data
    np.testing.assert_allclose(random, jnp.fft.ifftn(jnp.fft.fftn(random)).real)
    np.testing.assert_allclose(
        jnp.fft.fftn(random),
        jnp.fft.fftn(jnp.fft.ifftn(jnp.fft.fftn(random)).real),
    )
    # Run tests with cryojax.utils and random data
    np.testing.assert_allclose(
        ifftn(fftn(random)).real, jnp.fft.ifftn(jnp.fft.fftn(random)).real
    )
    np.testing.assert_allclose(random, ifftn(fftn(random)).real)
    np.testing.assert_allclose(fftn(random), fftn(ifftn(fftn(random))))
    # Run tests with an image
    np.testing.assert_allclose(image, ifftn(fftn(image)).real)
    # ... test zero mode separately
    np.testing.assert_allclose(
        fftn(image)[1:, 1:], fftn(ifftn(fftn(image)).real)[1:, 1:]
    )
    np.testing.assert_allclose(
        fftn(image)[0, 0], fftn(ifftn(fftn(image)).real)[0, 0], atol=1e-12
    )



from ._assembly import AbstractAssembly as AbstractAssembly
from ._helix import (
    compute_helical_lattice_positions as compute_helical_lattice_positions,
    compute_helical_lattice_rotations as compute_helical_lattice_rotations,
    Helix as Helix,
)



from ._scattering_potential import (
    AbstractScatteringPotential as AbstractScatteringPotential,
)
from ._voxel_potential import (
    AbstractFourierVoxelGridPotential as AbstractFourierVoxelGridPotential,
    AbstractVoxelPotential as AbstractVoxelPotential,
    build_real_space_voxels_from_atoms as build_real_space_voxels_from_atoms,
    evaluate_3d_atom_potential as evaluate_3d_atom_potential,
    evaluate_3d_real_space_gaussian as evaluate_3d_real_space_gaussian,
    FourierVoxelGridPotential as FourierVoxelGridPotential,
    FourierVoxelGridPotentialInterpolator as FourierVoxelGridPotentialInterpolator,
    RealVoxelCloudPotential as RealVoxelCloudPotential,
    RealVoxelGridPotential as RealVoxelGridPotential,
)



import numpy as np
from jaxtyping import Float, Int

from ._gemmi import (
    clean_gemmi_structure,
    extract_atom_positions_and_names,
    extract_gemmi_atoms,
)


def read_atoms_from_cif(
    path, i_model=0, clean=True, assemble=True
) -> tuple[Float[np.ndarray, "N 3"], Int[np.ndarray, " N"]]:
    """Read atomic positions and element names from a mmCIF file using Gemmi

    Parameters
    ----------
    path : string
        Path to mmCIF file.
    i_model : integer
        Optional, default: 0
        Index of the returned model in the Gemmi Structure.
    clean : bool
        Optional, default: True
        If True, use Gemmi remove_* methods to clean up structure.
    assemble: bool
        Optional, default: True
        If True, use Gemmi make_assembly to build biological object.

    Returns
    -------
    atom_positions: list of numpy arrays
        List of coordinates containing atomic positions
    atom_element_names: list of strings
        List of atomic element names
    """
    import gemmi

    cif_block = gemmi.cif.read(path)[0]
    structure = gemmi.make_structure_from_block(cif_block)
    if clean:
        structure = clean_gemmi_structure(structure)
    model = structure[i_model]
    if assemble:
        assembly = structure.assemblies[i_model]
        chain_naming = gemmi.HowToNameCopiedChain.AddNumber
        model = gemmi.make_assembly(assembly, model, chain_naming)
    atoms = extract_gemmi_atoms(model)
    atom_positions, atom_element_names = extract_atom_positions_and_names(atoms)
    return atom_positions, atom_element_names



"""
Representations of rigid body rotations and translations of 3D coordinate systems.
"""

from abc import abstractmethod
from functools import cached_property
from typing import overload
from typing_extensions import override

import equinox as eqx
import jax
import jax.numpy as jnp
from equinox import AbstractVar, field, Module
from jaxtyping import Array, Complex, Float

from ..rotations import SO3


class AbstractPose(Module, strict=True):
    """Base class for the image pose.

    Subclasses should choose a viewing convention,
    such as with Euler angles or Quaternions. In particular,

        1. Define angular coordinates as dataclass fields, along
           with other dataclass fields.

        2. Overwrite the `AbstractPose.rotation` property and
           `AbstractPose.from_rotation` class method.
    """

    offset_x_in_angstroms: AbstractVar[Float[Array, ""]]
    offset_y_in_angstroms: AbstractVar[Float[Array, ""]]
    offset_z_in_angstroms: AbstractVar[Float[Array, ""]]

    @overload
    def rotate_coordinates(
        self,
        volume_coordinates: Float[Array, "z_dim y_dim x_dim 3"],
        inverse: bool = False,
    ) -> Float[Array, "z_dim y_dim x_dim 3"]: ...

    @overload
    def rotate_coordinates(
        self, volume_coordinates: Float[Array, "size 3"], inverse: bool = False
    ) -> Float[Array, "size 3"]: ...

    def rotate_coordinates(
        self,
        volume_coordinates: (
            Float[Array, "z_dim y_dim x_dim 3"] | Float[Array, "size 3"]
        ),
        inverse: bool = False,
    ) -> Float[Array, "z_dim y_dim x_dim 3"] | Float[Array, "size 3"]:
        """Rotate coordinates from a particular convention."""
        rotation = self.rotation.inverse() if inverse else self.rotation
        if isinstance(volume_coordinates, Float[Array, "size 3"]):  # type: ignore
            rotated_volume_coordinates = jax.vmap(rotation.apply)(volume_coordinates)
        elif isinstance(volume_coordinates, Float[Array, "z_dim y_dim x_dim 3"]):  # type: ignore
            rotated_volume_coordinates = jax.vmap(jax.vmap(jax.vmap(rotation.apply)))(
                volume_coordinates
            )
        else:
            raise ValueError(
                "Coordinates must be a JAX array either of shape (N, 3) or "
                f"(N1, N2, N3, 3). Instead, got {volume_coordinates.shape} and type "
                f"{type(volume_coordinates)}."
            )
        return rotated_volume_coordinates

    def compute_shifts(
        self, frequency_grid_in_angstroms: Float[Array, "y_dim x_dim 2"]
    ) -> Complex[Array, "y_dim x_dim"]:
        """Compute the phase shifts from the in-plane translation,
        given a frequency grid coordinate system.
        """
        xy = self.offset_in_angstroms[0:2]
        return jnp.exp(
            -1.0j * (2 * jnp.pi * jnp.matmul(frequency_grid_in_angstroms, xy))
        )

    @cached_property
    def offset_in_angstroms(self) -> Float[Array, "3"]:
        """The translation vector, relative to the center of the in-plane
        (x, y) coordinates and relative to the configured defocus in the
        out-of-plane z coordinate.
        """
        return jnp.asarray(
            (
                self.offset_x_in_angstroms,
                self.offset_y_in_angstroms,
                self.offset_z_in_angstroms,
            )
        )

    @cached_property
    @abstractmethod
    def rotation(self) -> SO3:
        """Generate an `SO3` object."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_rotation(cls, rotation: SO3):
        """Construct an `AbstractPose` from an `SO3` object."""
        raise NotImplementedError

    @classmethod
    def from_rotation_and_translation(
        cls,
        rotation: SO3,
        offset_in_angstroms: Float[Array, "3"],
    ):
        """Construct an `AbstractPose` from an `AbstractRotation` object and a
        translation vector.
        """
        return eqx.tree_at(
            lambda self: (
                self.offset_x_in_angstroms,
                self.offset_y_in_angstroms,
                self.offset_z_in_angstroms,
            ),
            cls.from_rotation(rotation),
            (
                offset_in_angstroms[..., 0],
                offset_in_angstroms[..., 1],
                offset_in_angstroms[..., 2],
            ),
        )


class EulerAnglePose(AbstractPose, strict=True):
    r"""An `AbstractPose` represented by Euler angles.
    Angles are given in degrees, and the sequence of rotations is
    given by a zyz extrinsic rotations.
    """

    offset_x_in_angstroms: Float[Array, ""] = field(default=0.0, converter=jnp.asarray)
    offset_y_in_angstroms: Float[Array, ""] = field(default=0.0, converter=jnp.asarray)
    offset_z_in_angstroms: Float[Array, ""] = field(default=0.0, converter=jnp.asarray)

    view_phi: Float[Array, ""] = field(default=0.0, converter=jnp.asarray)
    view_theta: Float[Array, ""] = field(default=0.0, converter=jnp.asarray)
    view_psi: Float[Array, ""] = field(default=0.0, converter=jnp.asarray)

    @cached_property
    @override
    def rotation(self) -> SO3:
        """Generate a `SO3` object from a set of Euler angles."""
        phi, theta, psi = self.view_phi, self.view_theta, self.view_psi
        # Convert to radians.
        phi = jnp.deg2rad(phi)
        theta = jnp.deg2rad(theta)
        psi = jnp.deg2rad(psi)
        # Get sequence of rotations.
        R1, R2, R3 = (
            SO3.from_z_radians(phi),
            SO3.from_y_radians(theta),
            SO3.from_z_radians(psi),
        )
        return R3 @ R2 @ R1

    @override
    @classmethod
    def from_rotation(cls, rotation: SO3):
        view_phi, view_theta, view_psi = _convert_quaternion_to_euler_angles(
            rotation.wxyz,
            "zyz",
        )
        return cls(view_phi=view_phi, view_theta=view_theta, view_psi=view_psi)


EulerAnglePose.__init__.__doc__ = """**Arguments:**

- `offset_x_in_angstroms` : In-plane translation in x direction.
- `offset_y_in_angstroms` : In-plane translation in y direction.
- `offset_z_in_angstroms` : Out-of-plane translation in the z
                            direction. The translation is measured
                            relative to the configured defocus.
- `view_phi`: Angle to rotate about first rotation axis, which is the z axis.
- `view_theta`: Angle to rotate about second rotation axis, which is the y axis.
- `view_psi`: Angle to rotate about third rotation axis, which is the z axis.
"""


class QuaternionPose(AbstractPose, strict=True):
    """An `AbstractPose` represented by unit quaternions."""

    offset_x_in_angstroms: Float[Array, ""] = field(default=0.0, converter=jnp.asarray)
    offset_y_in_angstroms: Float[Array, ""] = field(default=0.0, converter=jnp.asarray)
    offset_z_in_angstroms: Float[Array, ""] = field(default=0.0, converter=jnp.asarray)

    wxyz: Float[Array, "4"] = field(default=(1.0, 0.0, 0.0, 0.0), converter=jnp.asarray)

    @cached_property
    @override
    def rotation(self) -> SO3:
        """Generate rotation from the unit quaternion."""
        # Generate SO3 object from unit quaternion
        R = SO3(wxyz=self.wxyz).normalize()
        return R

    @override
    @classmethod
    def from_rotation(cls, rotation: SO3):
        return cls(wxyz=rotation.wxyz)


QuaternionPose.__init__.__doc__ = r"""**Arguments:**

- `offset_x_in_angstroms` : In-plane translation in x direction.
- `offset_y_in_angstroms` : In-plane translation in y direction.
- `offset_z_in_angstroms` : Out-of-plane translation in the z
                            direction. The translation is measured
                            relative to the configured defocus.
- `wxyz`: The quaternion, represented as a vector $\mathbf{q} = (q_w, q_x, q_y, q_z)$.
"""


class AxisAnglePose(AbstractPose, strict=True):
    """An `AbstractPose` parameterized in the axis-angle representation.

    The axis-angle representation parameterizes elements of the so3 algebra,
    which are skew-symmetric matrices, with the euler vector. The magnitude
    of this vector is the angle, and the unit vector is the axis.

    In a `SO3` object, the euler vector is mapped to SO3 group elements using
    the matrix exponential.
    """

    offset_x_in_angstroms: Float[Array, ""] = field(default=0.0, converter=jnp.asarray)
    offset_y_in_angstroms: Float[Array, ""] = field(default=0.0, converter=jnp.asarray)
    offset_z_in_angstroms: Float[Array, ""] = field(default=0.0, converter=jnp.asarray)

    euler_vector: Float[Array, "3"] = field(
        default=(0.0, 0.0, 0.0), converter=jnp.asarray
    )

    @cached_property
    @override
    def rotation(self) -> SO3:
        """Generate rotation from an euler vector using the exponential map."""
        # Convert degrees to radians
        euler_vector = jnp.deg2rad(self.euler_vector)
        # Project the tangent vector onto the manifold with
        # the exponential map
        R = SO3.exp(euler_vector)
        return R

    @override
    @classmethod
    def from_rotation(cls, rotation: SO3):
        # Compute the euler vector from the logarithmic map
        euler_vector = jnp.rad2deg(rotation.log())
        return cls(euler_vector=euler_vector)


AxisAnglePose.__init__.__doc__ = r"""**Arguments:**

- `offset_x_in_angstroms` : In-plane translation in x direction.
- `offset_y_in_angstroms` : In-plane translation in y direction.
- `offset_z_in_angstroms` : Out-of-plane translation in the z
                            direction. The translation is measured
                            relative to the configured defocus.
- `euler_vector`: The axis-angle parameterization, represented as a
                  vector $\boldsymbol{\omega} = (\omega_x, \omega_y, \omega_z)$.
"""


def _convert_quaternion_to_euler_angles(
    wxyz: jax.Array, convention: str = "zyz"
) -> jax.Array:
    """Convert a quaternion to a sequence of euler angles about an extrinsic
    coordinate system.

    Adapted from https://github.com/chrisflesher/jax-scipy-spatial/.
    """
    if len(convention) != 3 or not all(
        [axis in ["x", "y", "z"] for axis in convention]
    ):
        raise ValueError(
            f"`convention` should be a string of three characters, each "
            f"of which is 'x', 'y', or 'z'. Instead, got '{convention}'"
        )
    if convention[0] == convention[1] or convention[1] == convention[2]:
        raise ValueError(
            f"`convention` cannot have axes repeating in a row. For example, "
            f"'xxy' or 'zzz' are not allowed. Got '{convention}'."
        )
    xyz_axis_to_array_axis = {"x": 0, "y": 1, "z": 2}
    axes = [xyz_axis_to_array_axis[axis] for axis in convention]
    xyzw = jnp.roll(wxyz, shift=-1)
    angle_first = 0
    angle_third = 2
    i = axes[0]
    j = axes[1]
    k = axes[2]
    symmetric = i == k
    k = jnp.where(symmetric, 3 - i - j, k)
    sign = jnp.array((i - j) * (j - k) * (k - i) // 2, dtype=xyzw.dtype)
    eps = 1e-7
    a = jnp.where(symmetric, xyzw[3], xyzw[3] - xyzw[j])
    b = jnp.where(symmetric, xyzw[i], xyzw[i] + xyzw[k] * sign)
    c = jnp.where(symmetric, xyzw[j], xyzw[j] + xyzw[3])
    d = jnp.where(symmetric, xyzw[k] * sign, xyzw[k] * sign - xyzw[i])
    angles = jnp.empty(3, dtype=xyzw.dtype)
    angles = angles.at[1].set(2 * jnp.arctan2(jnp.hypot(c, d), jnp.hypot(a, b)))
    case = jnp.where(jnp.abs(angles[1] - jnp.pi) <= eps, 2, 0)
    case = jnp.where(jnp.abs(angles[1]) <= eps, 1, case)
    half_sum = jnp.arctan2(b, a)
    half_diff = jnp.arctan2(d, c)
    angles = angles.at[0].set(
        jnp.where(case == 1, 2 * half_sum, 2 * half_diff * -1)
    )  # any degenerate case
    angles = angles.at[angle_first].set(
        jnp.where(case == 0, half_sum - half_diff, angles[angle_first])
    )
    angles = angles.at[angle_third].set(
        jnp.where(case == 0, half_sum + half_diff, angles[angle_third])
    )
    angles = angles.at[angle_third].set(
        jnp.where(symmetric, angles[angle_third], angles[angle_third] * sign)
    )
    angles = angles.at[1].set(jnp.where(symmetric, angles[1], angles[1] - jnp.pi / 2))
    angles = (angles + jnp.pi) % (2 * jnp.pi) - jnp.pi
    return -jnp.rad2deg(angles)



import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np
import pytest


@pytest.mark.parametrize("model", ["noisy_model"])
def test_compute_with_filters_and_masks(
    model, filtered_and_masked_model, request, filters, masks
):
    """Make sure that adding null filters and masks does not change output"""
    model = request.getfixturevalue(model)
    # Add null filters and masks
    null_mask = eqx.tree_at(lambda m: m.buffer, masks, jnp.asarray(1.0))
    null_filter = eqx.tree_at(lambda f: f.buffer, filters, jnp.asarray(1.0))
    where = lambda m: (m.filter, m.mask)
    model_with_null_mask = eqx.tree_at(
        where, filtered_and_masked_model, (None, null_mask)
    )
    model_with_null_filter = eqx.tree_at(
        where, filtered_and_masked_model, (null_filter, None)
    )
    model_with_null_filter_and_mask = eqx.tree_at(
        where,
        filtered_and_masked_model,
        (null_filter, null_mask),
    )
    # Compute images
    key = jax.random.PRNGKey(0)
    image = model.render()
    noisy_image = model.sample(key)
    # Check render
    np.testing.assert_allclose(model_with_null_mask.render(), image)
    np.testing.assert_allclose(model_with_null_filter.render(), image)
    np.testing.assert_allclose(model_with_null_filter_and_mask.render(), image)
    # Check sample
    np.testing.assert_allclose(model_with_null_mask.sample(key), noisy_image)
    np.testing.assert_allclose(model_with_null_filter.sample(key), noisy_image)
    np.testing.assert_allclose(model_with_null_filter_and_mask.sample(key), noisy_image)



"""
Implementation of transformations in local tangent spaces of lie groups.
This is a re-implementation of `jaxlie.manifold`.

This was necessary because `cryojax` has its own pattern for `jax`
transformations, using `equinox`. The `jaxlie.manifold.grad` and
`jaxlie.manifold.value_and_grad` are not compatible with the `cryojax`
wrappers for gradient transformations.
"""

import equinox as eqx
import jax
import jax.numpy as jnp
import jax.tree_util as jtu
from equinox import AbstractVar
from jaxtyping import Array, Float, PyTree

from ...rotations import AbstractMatrixLieGroup, SE3, SO3
from ...simulator import QuaternionPose
from ._transforms import AbstractParameterTransform


def _apply_update_with_lie_transform(u, p):
    if u is None:
        return p
    elif isinstance(u, AbstractLieGroupTransform):
        lie_group = type(u.group_element)
        local_tangent = u.transformed_parameter
        return eqx.tree_at(
            lambda p: p.group_element,
            p,
            p.group_element @ lie_group.exp(local_tangent),
        )
    else:
        return p + u


def _is_none_or_lie_transform(x):
    return x is None or isinstance(x, AbstractLieGroupTransform)


def apply_updates_with_lie_transform(model: PyTree, updates: PyTree) -> PyTree:
    """Modifed `eqx.apply_updates` to apply updates to a model
    with `AbstractLieGroupTransform`s.

    This assumes that `updates` are a prefix of `model`.
    """
    return jtu.tree_map(
        _apply_update_with_lie_transform,
        updates,
        model,
        is_leaf=_is_none_or_lie_transform,
    )


class AbstractLieGroupTransform(AbstractParameterTransform, strict=True):
    """An abstract base class for lie group transforms."""

    group_element: AbstractVar[AbstractMatrixLieGroup]


class SO3Transform(AbstractLieGroupTransform, strict=True):
    """This class transforms a quaternion to the local
    tangent space of the corresponding SO3 element.

    This class is based on the implementation in `jaxlie.manifold`.

    **Attributes:**

    - `transformed_parameter`: The local tangent vector.

    - `group_element`: The element of SO3.
    """

    transformed_parameter: Float[Array, "3"]
    group_element: SO3

    def __init__(self, wxyz: Float[Array, "4"]):
        """**Arguments:**

        - `wxyz`: A quaternion that parameterizes the SO3
                  group element.
        """
        local_tangent = jnp.zeros(3, dtype=float)
        self.transformed_parameter = local_tangent
        self.group_element = SO3(wxyz).normalize()

    def get(self) -> Float[Array, "4"]:
        """An implementation of the `jaxlie.manifold.rplus`."""
        local_tangent = self.transformed_parameter
        return (jax.lax.stop_gradient(self.group_element) @ SO3.exp(local_tangent)).wxyz


class SE3Transform(AbstractLieGroupTransform, strict=True):
    """This class transforms a `QuaternionPose` to the local
    tangent space of the corresponding SE3 element.

    This class is based on the implementation in `jaxlie.manifold`.

    **Attributes:**

    - `transformed_parameter`: The local tangent vector.

    - `group_element`: The element of SE3.
    """

    transformed_parameter: Float[Array, "6"]
    group_element: SE3

    def __init__(self, quaternion_pose: QuaternionPose):
        """**Arguments:**

        - `quaternion_pose`: A quaternion pose representation that parameterizes
                             the SE3 group element.
        """
        local_tangent = jnp.zeros(6, dtype=float)
        self.transformed_parameter = local_tangent
        self.group_element = SE3(
            rotation=quaternion_pose.rotation,
            xyz=quaternion_pose.offset_in_angstroms,
        )

    def get(self) -> Float[Array, "6"]:
        """An implementation of the `jaxlie.manifold.rplus`."""
        local_tangent = self.transformed_parameter
        group_element = jax.lax.stop_gradient(self.group_element) @ SE3.exp(
            local_tangent
        )
        return QuaternionPose.from_rotation_and_translation(
            group_element.rotation, group_element.xyz
        )



from ._coordinates import (
    AbstractCoordinates as AbstractCoordinates,
    cartesian_to_polar as cartesian_to_polar,
    CoordinateGrid as CoordinateGrid,
    CoordinateList as CoordinateList,
    FrequencyGrid as FrequencyGrid,
    FrequencySlice as FrequencySlice,
    make_coordinates as make_coordinates,
    make_frequencies as make_frequencies,
)



"""
Utilities for creating equinox filter_specs.
"""

from typing import Any, Callable, Optional, Sequence, Union

import equinox as eqx
import jax.tree_util as jtu
from jaxtyping import PyTree


def get_filter_spec(
    pytree: PyTree,
    where: Callable[[PyTree], Union[Any, Sequence[Any]]],
    *,
    inverse: bool = False,
    is_leaf: Optional[Callable[[Any], bool]] = None,
) -> PyTree[bool]:
    if not inverse:
        false_pytree = jtu.tree_map(lambda _: False, pytree)
        return eqx.tree_at(
            where, false_pytree, replace_fn=lambda _: True, is_leaf=is_leaf
        )
    else:
        true_pytree = jtu.tree_map(lambda _: True, pytree)
        return eqx.tree_at(
            where, true_pytree, replace_fn=lambda _: False, is_leaf=is_leaf
        )



import jax
import jax.random as jr
import numpy as np
import pytest


jax.config.update("jax_enable_x64", True)


@pytest.mark.parametrize("model", ["noisy_model"])
def test_jit(model, test_image, request):
    model = request.getfixturevalue(model)
    key = jr.PRNGKey(0)

    @jax.jit
    def compute_image(model, key):
        return model.sample(key)

    np.testing.assert_allclose(compute_image(model, key), model.sample(key))



"""
Implementation of operators on images in real-space.
"""

from abc import abstractmethod
from typing import overload
from typing_extensions import override

import jax.numpy as jnp
from equinox import field
from jaxtyping import Array, Float

from ...core import error_if_not_positive
from ._operator import AbstractImageOperator


class AbstractRealOperator(AbstractImageOperator, strict=True):
    """
    The base class for all real operators.

    By convention, operators should be defined to
    have units of inverse area (up to a scale factor).

    To create a subclass,

        1) Include the necessary parameters in
           the class definition.
        2) Overrwrite the ``__call__`` method.
    """

    @overload
    @abstractmethod
    def __call__(
        self, coords: Float[Array, "y_dim x_dim 2"]
    ) -> Float[Array, "y_dim x_dim"]: ...

    @overload
    @abstractmethod
    def __call__(
        self, coords: Float[Array, "z_dim y_dim x_dim 3"]
    ) -> Float[Array, "z_dim y_dim x_dim"]: ...

    @abstractmethod
    def __call__(  # pyright: ignore
        self,
        coords: Float[Array, "y_dim x_dim 2"] | Float[Array, "z_dim y_dim x_dim 3"],
    ) -> Float[Array, "y_dim x_dim"] | Float[Array, "z_dim y_dim x_dim"]:
        raise NotImplementedError


RealOperatorLike = AbstractRealOperator | AbstractImageOperator


class Gaussian2D(AbstractRealOperator, strict=True):
    r"""This operator represents a simple gaussian.
    Specifically, this is

    $$g(r) = \frac{\kappa}{2\pi \beta} \exp(- (r - r_0)^2 / (2 \beta))$$

    where $r^2 = x^2 + y^2$.
    """

    amplitude: Float[Array, ""] = field(default=1.0, converter=jnp.asarray)
    b_factor: Float[Array, ""] = field(default=1.0, converter=error_if_not_positive)
    offset: Float[Array, "2"] = field(default=(0.0, 0.0), converter=jnp.asarray)

    @override
    def __call__(
        self, coords: Float[Array, "y_dim x_dim 2"]
    ) -> Float[Array, "y_dim x_dim"]:
        r_sqr = jnp.sum((coords - self.offset) ** 2, axis=-1)
        scaling = (self.amplitude / jnp.sqrt(2 * jnp.pi * self.b_factor)) * jnp.exp(
            -0.5 * r_sqr / self.b_factor
        )
        return scaling


Gaussian2D.__init__.__doc__ = """**Arguments:**

- `amplitude`: The amplitude of the operator, equal to $\\kappa$
               in the above equation.
- `b_factor`: The variance of the gaussian, equal to $\\beta$
              in the above equation.
- `offset`: An offset to the origin, equal to $r_0$
            in the above equation.
"""
